from __future__ import annotations

import json
import os
import shlex
import shutil
import subprocess
import sys
from dataclasses import asdict

import pytest
from vunnel import result, workspace
from vunnel.providers import ubuntu
from vunnel.providers.ubuntu.parser import (
    CVEFile,
    Parser,
    Patch,
    check_merge,
    check_patch,
    map_parsed,
    parse_cve_file,
    parse_list,
    parse_multiline_keyvalue,
    parse_severity_from_priority,
    parse_simple_keyvalue,
    patch_states,
    Severity,
    ubuntu_version_names,
)


class TestUbuntuParser:
    _location_ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__), "test-fixtures"))
    _data_ = os.path.join(_location_, "example_ubuntu_cve")
    _weird_data_ = os.path.join(_location_, "weird_example_cve")
    _workspace_ = "/tmp/ubuntu"

    # @classmethod
    # def setUpClass(cls):
    #     os.makedirs(cls._workspace_, exist_ok=True)

    # @classmethod
    # def tearDownClass(cls):
    #     shutil.rmtree(cls._workspace_)

    def test_reference_parser(self):
        data = {
            """References:
            https://cve.mitre.org/1/cvename.cgi?somecve
            https://cve.mitre.org/2/cvename2.cgi
            """: [
                "https://cve.mitre.org/1/cvename.cgi?somecve",
                "https://cve.mitre.org/2/cvename2.cgi",
            ],
            """References:
            https://cve.mitre.org/1/cvename.cgi?somecve
            https://cve.mitre.org/2/cvename2.cgi

            """: [
                "https://cve.mitre.org/1/cvename.cgi?somecve",
                "https://cve.mitre.org/2/cvename2.cgi",
            ],
            """References:

            https://cve.mitre.org/1/cvename.cgi?somecve
            https://cve.mitre.org/2/cvename2.cgi

            """: [],
        }

        header = "References"

        for input, result in data.items():
            split_lines = input.splitlines()
            got = parse_list(header, split_lines)

            assert got == result

    def test_simple_newline_parser(self):
        data = {
            "Header: somevalue\n": "somevalue",
            "Header:   somevalue2     \n\n\n": "somevalue2",
        }

        header = "Header"

        for input, result in data.items():
            split_lines = input.splitlines()
            got = parse_simple_keyvalue(header, split_lines)
            assert got == result

    def test_parse_cve(self):
        with open(self._data_) as f:
            data = f.readlines()

        print("Parsing")
        result = parse_cve_file("CVE-2017-9996", data)
        print("Complete")

        expected = CVEFile.from_dict(
            {
                "Candidate": "CVE-2017-9996",
                # "PublicDate": dateutil.parser.parse("2017-06-28"),
                "References": [
                    "https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2017-9996",
                    "https://github.com/FFmpeg/FFmpeg/commit/1e42736b95065c69a7481d0cf55247024f54b660",
                    "https://github.com/FFmpeg/FFmpeg/commit/e1b60aad77c27ed5d4dfc11e5e6a05a38c70489d",
                    "https://bugs.chromium.org/p/oss-fuzz/issues/detail?id=1378",
                    "https://bugs.chromium.org/p/oss-fuzz/issues/detail?id=1427",
                ],
                "Description": "The cdxl_decode_frame function in libavcodec/cdxl.c in FFmpeg 2.8.x before 2.8.12, 3.0.x before 3.0.8, 3.1.x before 3.1.8, 3.2.x before 3.2.5, and 3.3.x before 3.3.1 does not exclude the CHUNKY format, which allows remote attackers to cause a denial of service (heap-based buffer overflow and application crash) or possibly have unspecified other impact via a crafted file.",
                # "Ubuntu-Description": "",
                "Priority": "medium",
                "patches": [
                    {
                        "distro": "upstream",
                        "package": "libav",
                        "status": "needs-triage",
                        "version": None,
                    },
                    {
                        "distro": "precise/esm",
                        "package": "libav",
                        "status": "DNE",
                        "version": None,
                    },
                    {
                        "distro": "trusty",
                        "package": "libav",
                        "status": "needed",
                        "version": None,
                    },
                    {
                        "distro": "vivid/ubuntu-core",
                        "package": "libav",
                        "status": "DNE",
                        "version": None,
                    },
                    {
                        "distro": "xenial",
                        "package": "libav",
                        "status": "DNE",
                        "version": None,
                    },
                    {
                        "distro": "yakkety",
                        "package": "libav",
                        "status": "DNE",
                        "version": None,
                    },
                    {
                        "distro": "zesty",
                        "package": "libav",
                        "status": "DNE",
                        "version": None,
                    },
                    {
                        "distro": "artful",
                        "package": "libav",
                        "status": "DNE",
                        "version": None,
                    },
                    {
                        "distro": "devel",
                        "package": "libav",
                        "status": "DNE",
                        "version": None,
                    },
                    {
                        "distro": "upstream",
                        "package": "ffmpeg",
                        "status": "released",
                        "version": "7:3.2.5-1",
                        "priority": "low",
                    },
                    {
                        "distro": "precise/esm",
                        "package": "ffmpeg",
                        "status": "DNE",
                        "version": None,
                        "priority": "low",
                    },
                    {
                        "distro": "trusty",
                        "package": "ffmpeg",
                        "status": "DNE",
                        "version": None,
                        "priority": "low",
                    },
                    {
                        "distro": "vivid/ubuntu-core",
                        "package": "ffmpeg",
                        "status": "DNE",
                        "version": None,
                        "priority": "low",
                    },
                    {
                        "distro": "xenial",
                        "package": "ffmpeg",
                        "status": "needed",
                        "version": None,
                        "priority": "low",
                    },
                    {
                        "distro": "yakkety",
                        "package": "ffmpeg",
                        "status": "ignored",
                        "version": "reached end-of-life",
                        "priority": "low",
                    },
                    {
                        "distro": "zesty",
                        "package": "ffmpeg",
                        "status": "needed",
                        "version": None,
                        "priority": "low",
                    },
                    {
                        "distro": "artful",
                        "package": "ffmpeg",
                        "status": "not-affected",
                        "priority": "low",
                        "version": "7:3.2.6-1",
                    },
                    {
                        "distro": "devel",
                        "package": "ffmpeg",
                        "status": "not-affected",
                        "priority": "low",
                        "version": "7:3.2.6-1",
                    },
                ],
            },
        )

        # No longer parsing by default
        # self.assertEqual(expected['PublicDate'], result['PublicDate'])
        assert expected.description == result.description
        self.maxDiff = None
        assert expected.patches == result.patches
        assert expected == result

    def test_non_header_patches(self):
        with open(self._weird_data_) as f:
            data = f.readlines()

        print("Parsing")
        result = parse_cve_file("CVE-2007-0255", data)
        print("Complete")

        print(result)

    def test_simple_multiline_parser(self):
        data = {
            """Description:
             a
             b
             c
             d
             e
             f
            """: "a b c d e f",
            """Description:
             ab
             bc


            """: "ab bc",
            """Description:

             a
             b

            """: "",
        }

        header = "Description"

        for input, result in data.items():
            split_lines = input.splitlines()
            got = parse_multiline_keyvalue(header, split_lines)
            assert got == result

    def test_mapper(self):
        with open(self._data_) as f:
            parsed = parse_cve_file("CVE-2017-9996", f.readlines())

        parsed.name = "CVE-TEST-123"
        vulns = map_parsed(parsed)
        for i in vulns:
            j = i.json()
            print(json.dumps(j))
            assert j != {"FixedIn": [{}, {}]}

    def test_checkers(self):
        check_data = [
            (Patch(distro="natty", status="released", version="1.1"), False),
            (Patch(distro="blah", status="released", version="1.1"), False),
            (Patch(distro="lucid", status="released", version="1.1"), False),
        ]

        # Try cross-product of states and releases
        for s, v in patch_states.items():
            for r in ubuntu_version_names:
                check_data.append((Patch(distro=r, status=s, version="testversion"), v))

        for data in check_data:
            assert check_patch(data[0]) == data[1]

    @pytest.mark.parametrize(
        ("patch", "expected"),
        [
            (Patch(distro="foo", package="bar", status="ignored ftw", version="end-of-life now but something else before"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="reached end-of-life"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="end-of-life"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="end-of-life, was needed"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="was pending now end-of-life"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="end_of_life"), False),
            (Patch(distro="foo", package="bar", status="ignored", version="bleddyend-of-lifeas we know"), False),
            (Patch(distro="foo", package="bar", status="ignored", version="end times of all life"), False),
            (Patch(distro="foo", package="bar", status="some-invalid-state", version="end-of-life"), False),
            (Patch(distro="foo", package="bar", status="ignored", version="oh so end-of-lifed"), False),
            (Patch(distro="foo", package="bar", status="ignored", version="end of standard support"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="out of standard support"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="end-of-standard-support"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="out-of-standard-support"), True),
            (Patch(distro="foo", package="bar", status="ignored", version="end of standard support, was needed"), True),
        ],
    )
    def test_check_merge(self, patch: Patch, expected: bool):
        assert check_merge(patch) == expected

    def test_reprocess_merged_cve(self, tmpdir):
        new_distro_patches = [
            {
                "distro": "madeup",
                "package": "mozjs38",
                "status": "DNE",
                "version": None,
            },
            {
                "distro": "madeup",
                "package": "mozjs52",
                "status": "needs-triage",
                "version": None,
            },
            {
                "distro": "madeup",
                "package": "mozjs60",
                "status": "ignored",
                "version": "end of life",
            },
        ]
        data = CVEFile.from_dict(
            {
                "patches": [
                    {
                        "distro": "trusty",
                        "package": "firefox",
                        "status": "ignored",
                        "version": "out of standard support",
                    },
                    {
                        "distro": "bionic",
                        "package": "firefox",
                        "status": "released",
                        "version": "82.0+build2-0ubuntu0.18.04.1",
                    },
                ],
                "Candidate": "CVE-0000-0000",
                "References": [
                    "https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-0000-0000",
                ],
                "Description": "blah blah blah",
                "Priority": "medium",
                "Name": "CVE-0000-0000",
                "ignored_patches": [
                    {"distro": "devel", "package": "mozjs52", "status": "DNE", "version": None},
                    *new_distro_patches,
                ],
                "git_last_processed_rev": "40a85b5b23bd905f8d5c6791d3f61108406ec372",
            },
        )
        ws = workspace.Workspace(tmpdir, "test")
        udp = Parser(workspace=ws, additional_versions={"madeup": "00.00"}, enable_rev_history=False)

        os.makedirs(udp.norm_workspace, exist_ok=True)

        cve_id = "CVE-0000-0000"
        cvs_file = os.path.join(udp.norm_workspace, cve_id)

        with open(cvs_file, "w") as fp:
            json.dump(asdict(data), fp)

        result = udp._reprocess_merged_cve(cve_id, cvs_file)
        assert result.patches == data.patches + [Patch(**p) for p in new_distro_patches]

    @pytest.mark.parametrize(
        ("cve", "expected_severity"),
        [
            (
                CVEFile(name="unset"),
                Severity.Unknown,
            ),
            (
                CVEFile(name="unknown", priority="unknown"),
                Severity.Unknown,
            ),
            (
                CVEFile(name="untriaged", priority="untriaged"),
                Severity.Unknown,
            ),
            (
                CVEFile(name="negligible", priority="negligible"),
                Severity.Negligible,
            ),
            (
                CVEFile(name="low", priority="low"),
                Severity.Low,
            ),
            (
                CVEFile(name="medium", priority="medium"),
                Severity.Medium,
            ),
            (
                CVEFile(name="high", priority="high"),
                Severity.High,
            ),
            (
                CVEFile(name="critical", priority="critical"),
                Severity.Critical,
            ),
        ],
    )
    def test_parse_severity_from_priority(self, cve: CVEFile, expected_severity: Severity):
        assert parse_severity_from_priority(cve) == expected_severity

    @pytest.mark.parametrize(
        ("cve", "error_type"),
        [
            (
                CVEFile(name="unset", priority="something-else"),
                AttributeError,
            ),
            (
                None,
                Exception,
            ),
        ],
    )
    def test_parse_severity_from_priority(self, cve: CVEFile, error_type: Exception):
        with pytest.raises(error_type):
            parse_severity_from_priority(cve)


@pytest.fixture()
def hydrate_git_repo(tmpdir, helpers):
    def run(cmd, **kwargs):
        subprocess.run(shlex.split(cmd), **kwargs, stderr=sys.stderr, stdout=sys.stdout)

    def apply(export_path):
        ws = workspace.Workspace(tmpdir, "ubuntu", create=True)
        repo_path = os.path.join(ws.input_path, "ubuntu-cve-tracker")

        shutil.rmtree(repo_path, ignore_errors=True)

        run("git init ubuntu-cve-tracker --initial-branch=main", cwd=ws.input_path)

        mock_data_path = helpers.local_dir(export_path)
        run("git fast-import", stdin=open(mock_data_path), cwd=repo_path)
        run("git checkout", cwd=repo_path)

        return tmpdir

    return apply


@pytest.mark.parametrize(
    ("mock_data_path", "expected_written_entries"),
    [
        ("test-fixtures/repo-fast-export", 42),
        # this is 6 records distributed across multiple distros:
        # └── results
        #     ├── ubuntu:14.04
        #     │   ├── cve-2019-17185.json
        #     │   ├── cve-2021-4204.json
        #     │   ├── cve-2022-20566.json
        #     │   ├── cve-2022-41859.json
        #     │   ├── cve-2022-41860.json
        #     │   └── cve-2022-41861.json
        #     ├── ubuntu:16.04
        #     │   ├── cve-2019-17185.json
        #     │   ├── cve-2021-4204.json
        #     │   ├── cve-2022-20566.json
        #     │   ├── cve-2022-41859.json
        #     │   ├── cve-2022-41860.json
        #     │   └── cve-2022-41861.json
        #     ├── ubuntu:18.04
        #     │   ├── cve-2019-17185.json
        #     │   ├── cve-2021-4204.json
        #     │   ├── cve-2022-20566.json
        #     │   ├── cve-2022-41859.json
        #     │   ├── cve-2022-41860.json
        #     │   └── cve-2022-41861.json
        #     ├── ubuntu:19.10
        #     │   └── cve-2019-17185.json
        #     ├── ubuntu:20.04
        #     │   ├── cve-2019-17185.json
        #     │   ├── cve-2021-4204.json
        #     │   ├── cve-2022-20566.json
        #     │   ├── cve-2022-41859.json
        #     │   ├── cve-2022-41860.json
        #     │   └── cve-2022-41861.json
        #     ├── ubuntu:20.10
        #     │   └── cve-2019-17185.json
        #     ├── ubuntu:21.04
        #     │   ├── cve-2019-17185.json
        #     │   └── cve-2021-4204.json
        #     ├── ubuntu:21.10
        #     │   ├── cve-2019-17185.json
        #     │   └── cve-2021-4204.json
        #     ├── ubuntu:22.04
        #     │   ├── cve-2019-17185.json
        #     │   ├── cve-2021-4204.json
        #     │   ├── cve-2022-20566.json
        #     │   ├── cve-2022-41859.json
        #     │   ├── cve-2022-41860.json
        #     │   └── cve-2022-41861.json
        #     └── ubuntu:22.10
        #         ├── cve-2019-17185.json
        #         ├── cve-2021-4204.json
        #         ├── cve-2022-20566.json
        #         ├── cve-2022-41859.json
        #         ├── cve-2022-41860.json
        #         └── cve-2022-41861.json
    ],
)
def test_provider_schema(helpers, mock_data_path, hydrate_git_repo, expected_written_entries, mocker):
    path = hydrate_git_repo(mock_data_path)

    c = ubuntu.Config()
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = ubuntu.Provider(root=path, config=c)
    p.parser.git_wrapper.init_repo = mocker.Mock()
    p.update(None)

    ws = helpers.provider_workspace_helper("ubuntu", create=False)

    assert expected_written_entries == ws.num_result_entries()
    assert ws.result_schemas_valid(require_entries=expected_written_entries > 0)


def test_provider_via_snapshot(helpers, hydrate_git_repo, mocker):
    path = hydrate_git_repo("test-fixtures/repo-fast-export")

    c = ubuntu.Config()
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = ubuntu.Provider(root=path, config=c)
    p.parser.git_wrapper.init_repo = mocker.Mock()
    p.update(None)

    ws = helpers.provider_workspace_helper("ubuntu", create=False)

    ws.assert_result_snapshots()



from __future__ import annotations

import json
import os
import shutil

import pytest
from vunnel import result, workspace
from vunnel.providers.minimos import Config, Provider, parser
from vunnel.providers.minimos.parser import Parser


class TestParser:
    @pytest.fixture()
    def mock_raw_data(self, helpers):
        """
        Returns stringified version of the following json

        ---
        """
        data = {
            "apkurl": "{{urlprefix}}/{{reponame}}/{{arch}}/{{pkg.name}}-{{pkg.ver}}.apk",
            "archs": ["x86_64"],
            "reponame": "os",
            "urlprefix": "https://packages.mini.dev",
            "packages": [
                {
                    "pkg": {
                        "name": "binutils",
                        "secfixes": {
                            "2.39-r1": ["CVE-2022-38126"],
                            "2.39-r2": ["CVE-2022-38533"],
                            "2.39-r3": ["CVE-2022-38128"],
                        },
                    },
                },
                {
                    "pkg": {
                        "name": "brotli",
                        "secfixes": {"1.0.9-r0": ["CVE-2020-8927"]},
                    },
                },
                {
                    "pkg": {
                        "name": "busybox",
                        "secfixes": {"1.35.0-r3": ["CVE-2022-28391", "CVE-2022-30065"]},
                    },
                },
                {"pkg": {"name": "coreutils", "secfixes": {"0": ["CVE-2016-2781"]}}},
                {"pkg": {"name": "cups", "secfixes": {"2.4.2-r0": ["CVE-2022-26691"]}}},
                {
                    "pkg": {
                        "name": "dbus",
                        "secfixes": {
                            "1.14.4-r0": [
                                "CVE-2022-42010",
                                "CVE-2022-42011",
                                "CVE-2022-42012",
                            ],
                        },
                    },
                },
                {
                    "pkg": {
                        "name": "apko",
                        "secfixes": {
                            "0": ["CVE-2023-45283", "CVE-2023-45284", "GHSA-jq35-85cj-fj4p"],
                            "0.10.0-r6": ["CVE-2023-39325", "CVE-2023-3978"],
                            "0.7.3-r1": ["CVE-2023-28840", "CVE-2023-28841", "CVE-2023-28842"],
                            "0.8.0-r1": ["CVE-2023-30551"],
                        },
                    },
                },
            ],
        }

        return json.dumps(data)

    @pytest.fixture()
    def mock_parsed_data(self):
        """
        Returns the parsed output generated by AlpineDataProvider._load() for the mock_raw_data

        :return:
        """
        release = "rolling"
        dbtype_data_dict = {
            "apkurl": "{{urlprefix}}/{{reponame}}/{{arch}}/{{pkg.name}}-{{pkg.ver}}.apk",
            "archs": ["x86_64"],
            "reponame": "os",
            "urlprefix": "https://packages.mini.dev",
            "packages": [
                {
                    "pkg": {
                        "name": "binutils",
                        "secfixes": {
                            "2.39-r1": ["CVE-2022-38126"],
                            "2.39-r2": ["CVE-2022-38533"],
                            "2.39-r3": ["CVE-2022-38128"],
                        },
                    },
                },
                {
                    "pkg": {
                        "name": "brotli",
                        "secfixes": {"1.0.9-r0": ["CVE-2020-8927"]},
                    },
                },
                {
                    "pkg": {
                        "name": "busybox",
                        "secfixes": {"1.35.0-r3": ["CVE-2022-28391", "CVE-2022-30065"]},
                    },
                },
                {
                    "pkg": {
                        "name": "coreutils",
                        "secfixes": {"0": ["CVE-2016-2781"]},
                    },
                },
                {
                    "pkg": {
                        "name": "cups",
                        "secfixes": {"2.4.2-r0": ["CVE-2022-26691"]},
                    },
                },
                {
                    "pkg": {
                        "name": "dbus",
                        "secfixes": {
                            "1.14.4-r0": [
                                "CVE-2022-42010",
                                "CVE-2022-42011",
                                "CVE-2022-42012",
                            ],
                        },
                    },
                },
                {
                    "pkg": {
                        "name": "apko",
                        "secfixes": {
                            "0": ["CVE-2023-45283", "CVE-2023-45284", "GHSA-jq35-85cj-fj4p"],
                            "0.10.0-r6": ["CVE-2023-39325", "CVE-2023-3978"],
                            "0.7.3-r1": ["CVE-2023-28840", "CVE-2023-28841", "CVE-2023-28842"],
                            "0.8.0-r1": ["CVE-2023-30551"],
                        },
                    },
                },
            ],
        }
        return release, dbtype_data_dict

    def test_load(self, mock_raw_data, tmpdir):
        p = Parser(
            workspace=workspace.Workspace(tmpdir, "test", create=True),
            url="https://packages.mini.dev/advisories/secdb/security.json",
            namespace="minimos",
        )

        os.makedirs(p.secdb_dir_path, exist_ok=True)
        b = os.path.join(p.secdb_dir_path, "security.json")
        with open(b, "w") as fp:
            fp.write(mock_raw_data)

        counter = 0
        for release, dbtype_data_dict in p._load():
            counter += 1
            assert release == "rolling"
            assert isinstance(dbtype_data_dict, dict)
            assert "packages" in dbtype_data_dict

        assert counter == 1

    def test_normalize(self, mock_parsed_data, tmpdir):
        p = Parser(
            workspace=workspace.Workspace(tmpdir, "test", create=True),
            url="https://packages.mini.dev/advisories/secdb/security.json",
            namespace="minimos",
        )
        release = mock_parsed_data[0]
        dbtype_data_dict = mock_parsed_data[1]

        vuln_records = p._normalize(release, dbtype_data_dict)
        assert len(vuln_records) > 0
        assert all("Vulnerability" in x for x in vuln_records.values())
        assert sorted(vuln_records.keys()) == sorted(
            [
                "CVE-2016-2781",
                "CVE-2020-8927",
                "CVE-2022-26691",
                "CVE-2022-28391",
                "CVE-2022-30065",
                "CVE-2022-38126",
                "CVE-2022-38128",
                "CVE-2022-38533",
                "CVE-2022-42010",
                "CVE-2022-42011",
                "CVE-2022-42012",
                "CVE-2023-28840",
                "CVE-2023-28841",
                "CVE-2023-28842",
                "CVE-2023-30551",
                "CVE-2023-39325",
                "CVE-2023-3978",
                "CVE-2023-45283",
                "CVE-2023-45284",
                "GHSA-jq35-85cj-fj4p",
            ],
        )


def test_provider_schema(helpers, disable_get_requests):
    workspace = helpers.provider_workspace_helper(
        name=Provider.name(),
        input_fixture="test-fixtures/input",
    )
    c = Config()
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(root=workspace.root, config=c)

    p.update(None)

    assert workspace.num_result_entries() == 59
    assert workspace.result_schemas_valid(require_entries=True)


def test_provider_via_snapshot(helpers, disable_get_requests, monkeypatch):
    workspace = helpers.provider_workspace_helper(
        name=Provider.name(),
        input_fixture="test-fixtures/input",
    )

    c = Config()
    # keep all of the default values for the result store, but override the strategy
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(
        root=workspace.root,
        config=c,
    )

    p.update(None)

    workspace.assert_result_snapshots()



from __future__ import annotations

import bz2
import copy
import gzip
import logging
import os
import re
from typing import TYPE_CHECKING

import defusedxml.ElementTree as ET

from vunnel.utils.vulnerability import vulnerability_element

if TYPE_CHECKING:
    from collections.abc import Callable

logger = logging.getLogger("oval-parser")


class Config:
    """
    Capture regular expressions, xpath queries and other driver specific configuration
    in an instance of this class

    """

    # regexes
    tag_pattern: re.Pattern | None = None
    ns_pattern: re.Pattern | None = None
    is_installed_pattern: re.Pattern | None = None
    pkg_version_pattern: re.Pattern | None = None
    pkg_module_pattern: re.Pattern | None = None
    signed_with_pattern: re.Pattern | None = None
    platform_version_pattern: re.Pattern | None = None

    # xpath queries
    title_xpath_query: str | None = None
    severity_xpath_query: str | None = None
    platform_xpath_query: str | None = None
    date_issued_xpath_query: str | None = None
    date_updated_xpath_query: str | None = None
    description_xpath_query: str | None = None
    sa_ref_xpath_query: str | None = None
    cve_xpath_query: str | None = None
    criteria_xpath_query: str | None = None
    criterion_xpath_query: str | None = None

    # maps
    severity_dict: dict[str, str] | None = None

    # string formats
    ns_format: str | None = None


def get_opener(filename: str) -> Callable:
    if filename.endswith(".gz"):
        return gzip.open
    if filename.endswith(".bz2"):
        return bz2.open
    return open


def parse(dest_file: str, config: Config, vuln_dict: dict | None = None):
    """
    Parse the oval file and return a dictionary with tuple (ID, namespace) as the key
    and tuple (version, vulnerability-dictionary) as the value
    :param dest_file: path object to oval file to be parsed
    :param config: configuration for parsing oval file
    :return:
    """
    if not isinstance(config, Config):
        logger.warning("Invalid config found, expected an instance of Config class")
        raise TypeError("Invalid config")

    logger.debug("Parsing {}".format(dest_file))  # noqa: UP032, G001

    if not vuln_dict:
        vuln_dict = {}

    if os.path.exists(dest_file):
        processing = False
        opener = get_opener(dest_file)

        with opener(dest_file, "rb") as f:  # noqa: F841
            for event, element in ET.iterparse(dest_file, events=("start", "end")):
                # gather definition
                if event == "start" and re.search(config.tag_pattern, element.tag).group(1) == "definition":
                    processing = True
                elif event == "end" and re.search(config.tag_pattern, element.tag).group(1) == "definition":
                    try:
                        _process_definition(element, vuln_dict, config)
                    except Exception:
                        logger.exception("Error parsing oval record. Logging error and continuing")
                    finally:
                        processing = False

                if not processing and event == "end":
                    # print('Clearing element: {} post event: {}'.format(re.search(tag_pattern, element.tag).group(1), event))
                    element.clear()

                # bail after definitions
                if event == "end" and re.search(config.tag_pattern, element.tag).group(1) == "definitions":
                    # print('Stopped parsing')
                    break
    else:
        logger.warning(f"{dest_file} not found, returning empty results")

    return vuln_dict


def _parse_description(def_element, oval_ns, config: Config) -> str:
    try:
        description = def_element.find(config.description_xpath_query.format(oval_ns)).text.strip()
    except (AttributeError, ET.ParseError):
        description = ""
    return description


def _parse_severity(def_element, oval_ns, vuln_id: str, config: Config) -> str:
    try:
        severity = config.severity_dict.get(def_element.find(config.severity_xpath_query.format(oval_ns)).text.lower())
    except (AttributeError, ET.ParseError):
        logger.debug(f"Unable to parse severity for {vuln_id}, defaulting to Unknown")
        severity = "Unknown"
    return severity


def _process_definition(def_element, vuln_dict, config: Config):  # noqa: PLR0912
    oval_ns = re.search(config.ns_pattern, def_element.tag).group(1)

    def_version = def_element.attrib["version"]
    title = def_element.find(config.title_xpath_query.format(oval_ns)).text
    name = title[: title.index(": ")].strip()
    description = _parse_description(def_element, oval_ns, config)
    severity = _parse_severity(def_element, oval_ns, name, config)
    issued = def_element.find(config.date_issued_xpath_query.format(oval_ns)).attrib["date"]
    # check for xpath query first since oracle does not provide this and its not initialized in the config
    if config.date_updated_xpath_query:  # noqa: SIM108
        updated = def_element.find(config.date_updated_xpath_query.format(oval_ns)).attrib["date"]
    else:
        updated = None
    rhsa_ref = def_element.find(config.sa_ref_xpath_query.format(oval_ns))
    ref_id = rhsa_ref.attrib["ref_id"]
    link = rhsa_ref.attrib["ref_url"]

    cves = []
    for cve in def_element.iterfind(config.cve_xpath_query.format(oval_ns)):
        if "cvss2" in cve.attrib:
            cves.append(
                {
                    "Name": cve.text,
                    "Link": cve.attrib["href"],
                    "cvss2": cve.attrib["cvss2"],
                },
            )
        else:
            cves.append({"Name": cve.text, "Link": cve.attrib["href"]})

    ns_pkgs_dict = _process_criteria(def_element, oval_ns, config)

    for platform_element in def_element.iterfind(config.platform_xpath_query.format(oval_ns)):
        v = copy.deepcopy(vulnerability_element)
        ns_name = config.ns_format.format(re.search(config.platform_version_pattern, platform_element.text).group(1))

        v["Vulnerability"]["NamespaceName"] = ns_name
        v["Vulnerability"]["Severity"] = severity or ""
        v["Vulnerability"]["Metadata"] = {"Issued": issued, "Updated": updated, "RefId": ref_id} if updated else {"Issued": issued, "RefId": ref_id}
        v["Vulnerability"]["Name"] = name
        v["Vulnerability"]["Link"] = link
        v["Vulnerability"]["Description"] = description

        if cves:
            v["Vulnerability"]["Metadata"]["CVE"] = cves

        if ns_pkgs_dict and ns_name in ns_pkgs_dict:
            v["Vulnerability"]["FixedIn"] = [
                {
                    "Name": x[0],
                    "Version": x[1],
                    "Module": x[2],
                    "VersionFormat": "rpm",  # hard code version format for now
                    "NamespaceName": ns_name,
                }
                for x in ns_pkgs_dict[ns_name]
            ]
        else:
            logger.warning(f"No affected packages found for {name}, this is unusual")

        # handle duplicates using version version attribute of definition element
        if (name, ns_name) in vuln_dict:
            existing_version, _ = vuln_dict[(name, ns_name)]
            logger.debug(
                "Found an existing record for {} under {}. Version attribute of definition oval element: existing: {}, new: {}".format(  # noqa: UP032, G001
                    name,
                    ns_name,
                    existing_version,
                    def_version,
                ),
            )
            # lexicographic comparison of versions to choose which vulnerability record wins
            if def_version > existing_version:
                # Replacing existing record with new one'
                vuln_dict[(name, ns_name)] = (def_version, v)
            else:
                # Existing record stays
                pass
        else:
            vuln_dict[(name, ns_name)] = (def_version, v)


def _process_criteria(element_a, oval_ns, config: Config):
    """
    Parse and return a dict containing namespace mapped to a set of (package, version) tuples
    :param element_a: outermost criteria element within a definition element
    :param oval_ns: namespace URL of the oval
    :return:
    """
    criteria_element = element_a.find(config.criteria_xpath_query.format(oval_ns))
    groups = []
    ns_pkgs_dict = {}

    if criteria_element.attrib["operator"].lower() == "or":
        for child in list(criteria_element):
            groups.append(_get_all_criterion(child, oval_ns, config))
    else:
        groups.append(_get_all_criterion(criteria_element, oval_ns, config))

    for group in groups:
        if not group:
            # logger.debug('Parsed group for one or more criterion is empty, skipping')
            continue  # bail out of processing the group if its empty

        # Find the first platform version string in the returned list
        ns_name, ns_module = next((x for x in group if not isinstance(x, tuple)), None)

        if ns_name:  # proceed only if a platform is found
            # Filter out duplicate (package, version) tuples
            ns_pkgs_dict[ns_name] = {tuple(list(x) + [ns_module]) for x in group if isinstance(x, tuple)}  # noqa: RUF005
        else:
            # logger.debug('Namespace for the criteria not found, ignoring criteria')
            continue  # ignore this group of conditions if namespace is not found

    return ns_pkgs_dict


def _get_all_criterion(element_b, oval_ns, config: Config):
    """
    Search for all the criterion elements under the given criteria element and
    parse contents into a list. Returned list may contain tuples and or simple strings.
    Package name and version found in the comment of a criterion element is represented by a tuple.
    Platform version found in the comment of a criterion element is represented by a simple string.
    :param element_b: criteria element
    :param oval_ns: namespace URL of the oval
    :return:
    """
    collectibles = []
    final_ns_name = None
    final_module_name = None
    ns_name = None
    module_name = None
    for criterion in element_b.iterfind(config.criterion_xpath_query.format(oval_ns)):
        if re.search(config.pkg_version_pattern, criterion.attrib["comment"]):
            pkg, version = re.search(config.pkg_version_pattern, criterion.attrib["comment"]).groups()
            collectibles.append((pkg, version))
        elif re.search(config.is_installed_pattern, criterion.attrib["comment"]):
            ns_name = config.ns_format.format(re.search(config.is_installed_pattern, criterion.attrib["comment"]).group(1))
        elif re.search(config.pkg_module_pattern, criterion.attrib["comment"]):
            module_name = re.search(config.pkg_module_pattern, criterion.attrib["comment"]).group(1)
        if ns_name:
            final_ns_name = ns_name
        if module_name:
            final_module_name = module_name
    if final_ns_name:
        collectibles.append([final_ns_name, final_module_name])
    return collectibles



import shutil
from unittest.mock import patch

from vunnel import result
from vunnel.providers.alma import Config, Provider
from vunnel.providers.alma.parser import Parser


@patch("vunnel.providers.alma.git.GitWrapper.clone_repo")
@patch("vunnel.providers.alma.git.GitWrapper.delete_repo")
def test_provider_schema(mock_git_delete, mock_git_clone, helpers):
    mock_git_clone.return_value = None
    mock_git_delete.return_value = None
    workspace = helpers.provider_workspace_helper(name=Provider.name())
    c = Config()
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(root=workspace.root, config=c)
    mock_data_path = helpers.local_dir("test-fixtures")
    shutil.copytree(mock_data_path, workspace.input_dir, dirs_exist_ok=True)
    p.update(None)

    assert 4 == workspace.num_result_entries()
    assert workspace.result_schemas_valid(require_entries=True)

@patch("vunnel.providers.alma.git.GitWrapper.clone_repo")
@patch("vunnel.providers.alma.git.GitWrapper.delete_repo")
def test_parser(mock_git_delete, mock_git_clone, helpers):
    mock_git_clone.return_value = None
    mock_git_delete.return_value = None
    workspace = helpers.provider_workspace_helper(name=Provider.name())
    mock_data_path = helpers.local_dir("test-fixtures")
    shutil.copytree(mock_data_path, workspace.input_dir, dirs_exist_ok=True)
    parser = Parser(ws=workspace, logger=None)
    vuln_tuples = list(parser.get())
    assert len(vuln_tuples) == 4
    assert vuln_tuples[0][0] == "almalinux8/ALBA-2021:4378"
    assert vuln_tuples[0][1] == "1.7.0"
    assert vuln_tuples[1][0] == "almalinux8/ALSA-2023:4520"
    assert vuln_tuples[1][1] == "1.7.0"
    assert vuln_tuples[2][0] == "almalinux9/ALSA-2022:8194"
    assert vuln_tuples[2][1] == "1.7.0"
    assert vuln_tuples[3][0] == "almalinux9/ALSA-2024:2433"
    assert vuln_tuples[3][1] == "1.7.0"



from __future__ import annotations

import datetime  # noqa: TC003
import logging
import os
import shutil
import sqlite3
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Optional

import orjson
import xxhash
from mashumaro.mixins.dict import DataClassDictMixin

from vunnel import schema as schema_def
from vunnel import utils
from vunnel.utils import hasher

if TYPE_CHECKING:
    import types
    from collections.abc import Generator

METADATA_FILENAME = "metadata.json"
CHECKSUM_LISTING_FILENAME = "checksums"


@dataclass
class File:
    digest: str
    path: str
    algorithm: str = "xxh64"


@dataclass
class State(DataClassDictMixin):
    provider: str
    urls: list[str]
    store: str
    timestamp: datetime.datetime
    version: int = 1
    distribution_version: int = 1
    processor: str | None = None
    listing: Optional[File] = None  # noqa:UP007  # why use Optional? mashumaro does not support this on python 3.9
    schema: schema_def.Schema = field(default_factory=schema_def.ProviderStateSchema)
    stale: bool = False

    def __post_init__(self) -> None:
        if not self.processor:
            metadata: types.ModuleType
            package_name = "vunnel"
            version = "dev"
            try:
                from importlib import metadata
            except ImportError:
                # Python < 3.8
                import importlib_metadata as metadata

            try:
                if not metadata:
                    raise metadata.PackageNotFoundError
                version = metadata.version(package_name)
            except metadata.PackageNotFoundError:
                version = "unknown"

            self.processor = f"{package_name}@{version}"

    @staticmethod
    def read(root: str) -> State:
        metadata_path = os.path.join(root, METADATA_FILENAME)
        with open(metadata_path, encoding="utf-8") as f:
            return State.from_dict(orjson.loads(f.read()))

    def write(self, root: str, results: str, update_listing: bool = True) -> str:
        metadata_path = os.path.join(root, METADATA_FILENAME)

        if update_listing:
            listing_path = os.path.join(root, CHECKSUM_LISTING_FILENAME)
            if self.listing:
                listing_path = os.path.join(root, self.listing.path)

            if os.path.exists(listing_path):
                os.remove(listing_path)

            # why not include the file listing in the metadata file?
            # because in some cases there is a lot of data and it's easier to stream
            # the results to a tab-delimited file than include in the metadata file
            self.listing = File(
                digest=write_file_listing(listing_path, results),
                algorithm="xxh64",
                path=os.path.basename(listing_path),  # may have been overridden, keep value
            )

        with open(metadata_path, "wb") as f:
            f.write(orjson.dumps(asdict(self), option=orjson.OPT_INDENT_2))

        return metadata_path

    def result_files(self, root: str) -> Generator[File, File, None]:
        if self.listing:
            full_path = os.path.join(root, self.listing.path)
            with open(full_path) as f:
                for digest, filepath in (line.split() for line in f.readlines()):
                    yield File(digest=digest, path=filepath, algorithm=self.listing.algorithm)

    def result_count(self, root: str) -> int:
        count = 0
        if self.listing:
            full_path = os.path.join(root, self.listing.path)
            with open(full_path) as f:
                for _digest, filepath in (line.split() for line in f.readlines()):
                    if filepath.endswith(".db"):
                        # open up the sqlite db and count the records in the "results" table
                        with sqlite3.connect(os.path.join(root, filepath)) as db:
                            count += db.execute("SELECT COUNT(*) FROM results").fetchone()[0]
                    else:
                        count += 1

        return count


class Workspace:
    def __init__(self, root: str, name: str, create: bool = False, logger: logging.Logger | None = None):
        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger
        self._root = root
        self.name = name

        if create:
            self.create()

    @property
    def path(self) -> str:
        return os.path.join(self._root, self.name)

    @property
    def results_path(self) -> str:
        return os.path.join(self.path, "results")

    @property
    def metadata_path(self) -> str:
        return os.path.join(self.path, METADATA_FILENAME)

    @property
    def checksums_path(self) -> str:
        return os.path.join(self.path, CHECKSUM_LISTING_FILENAME)

    @property
    def input_path(self) -> str:
        return os.path.join(self.path, "input")

    def create(self) -> None:
        if not os.path.exists(self.input_path):
            self.logger.debug(f"creating input workspace {self.input_path!r}")
            os.makedirs(self.input_path)
        else:
            self.logger.debug(f"using existing input workspace {self.input_path!r}")

        if not os.path.exists(self.results_path):
            self.logger.debug(f"creating results workspace {self.results_path!r}")
            os.makedirs(self.results_path)
        else:
            self.logger.debug(f"using existing results workspace {self.results_path!r}")

    def clear(self) -> None:
        self.clear_input()
        self.clear_results()
        self._clear_metadata()

    def _clear_metadata(self) -> None:
        utils.silent_remove(os.path.join(self.path, METADATA_FILENAME))
        utils.silent_remove(os.path.join(self.path, CHECKSUM_LISTING_FILENAME))

    def clear_results(self, recreate_results_dir: bool = True) -> None:
        if os.path.exists(self.results_path):
            self.logger.debug("clearing existing results")
            shutil.rmtree(self.results_path)
            if recreate_results_dir:
                os.makedirs(self.results_path, exist_ok=True)

        try:
            current_state = State.read(root=self.path)
            current_listing = None
            if current_state.listing:
                current_listing = current_state.listing.path
            current_state.listing = None
            current_state.write(self.path, self.results_path)
            if current_listing:
                utils.silent_remove(current_listing)
        except FileNotFoundError:
            pass

    def clear_input(self) -> None:
        if os.path.exists(self.input_path):
            self.logger.debug("clearing existing input")
            shutil.rmtree(self.input_path)
            os.makedirs(self.input_path, exist_ok=True)

    def record_state(  # noqa: PLR0913
        self,
        version: int,
        distribution_version: int,
        timestamp: datetime.datetime,
        urls: list[str],
        store: str,
        stale: bool = False,
    ) -> None:
        try:
            current_state = State.read(root=self.path)
        except FileNotFoundError:
            current_state = None

        if current_state:
            if not urls:
                urls = current_state.urls
            if not timestamp:
                timestamp = current_state.timestamp

        self.logger.info("recording workspace state")

        state = State(
            provider=self.name,
            version=version,
            distribution_version=distribution_version,
            urls=urls,
            store=store,
            timestamp=timestamp,
            stale=stale,
        )
        metadata_path = state.write(self.path, self.results_path)

        self.logger.debug(f"wrote workspace state to {metadata_path}")

    def state(self) -> State:
        return State.read(self.path)

    def validate_checksums(self) -> None:
        state = State.read(self.path)
        if not state.listing:
            raise RuntimeError("no file listing found in workspace state")

        full_path = os.path.join(self.path, state.listing.path)

        # ensure the checksums file itself is not modified
        if state.listing.digest != hasher.Method.XXH64.digest(full_path, label=False):
            raise RuntimeError(f"file {full_path!r} has been modified")

        # validate the checksums in the listing file
        with open(full_path) as f:
            for line in f.readlines():
                digest, path = line.split()
                full_path = os.path.join(self.path, path)
                if not os.path.exists(full_path):
                    raise RuntimeError(f"file {full_path!r} does not exist")

                if digest != hasher.Method.XXH64.digest(full_path, label=False):
                    raise RuntimeError(f"file {full_path!r} has been modified")

    def replace_results(self, temp_workspace: Workspace) -> None:
        self.logger.info(f"replacing results in {self.path!r} with results from {temp_workspace.path!r}")
        self.clear_results(recreate_results_dir=False)
        shutil.move(temp_workspace.results_path, self.path)
        self._clear_metadata()
        shutil.move(temp_workspace.metadata_path, self.metadata_path)
        shutil.move(temp_workspace.checksums_path, self.checksums_path)
        state = self.state()
        state.stale = True
        self.record_state(state.version, state.distribution_version, state.timestamp, state.urls, state.store, True)


def write_file_listing(output_file: str, path: str) -> str:
    listing_hasher = xxhash.xxh64()

    with open(output_file, "w", encoding="utf-8") as f:
        for root, dirs, files in os.walk(path):
            dirs.sort()  # sort the existing list that os.walk generator continues to reference
            for file in sorted(files):
                full_path = os.path.join(root, file)
                path_relative_to_results = os.path.relpath(full_path, path)
                path_relative_to_workspace = os.path.join(os.path.basename(path), path_relative_to_results)

                contents = f"{hasher.Method.XXH64.digest(full_path, label=False)}  {path_relative_to_workspace}\n"
                listing_hasher.update(contents.encode("utf-8"))

                f.write(contents)

    return listing_hasher.hexdigest()



from __future__ import annotations

import logging
import pytest
import requests
from unittest.mock import patch, MagicMock, call
from vunnel.utils import http_wrapper as http


class TestGetRequests:
    @pytest.fixture()
    def mock_logger(self):
        logger = logging.getLogger("test-http-utils")
        return MagicMock(logger, autospec=True)

    @pytest.fixture()
    def error_response(self):
        mock_response = MagicMock()
        mock_response.raise_for_status = MagicMock()
        mock_response.raise_for_status.side_effect = requests.HTTPError("HTTP ERROR")
        return mock_response

    @pytest.fixture()
    def success_response(self):
        response = MagicMock()
        response.raise_for_status = MagicMock()
        response.raise_for_status.side_effect = None
        response.status_code = 200
        return response

    @patch("time.sleep")
    @patch("requests.get")
    def test_raises_when_out_of_retries(self, mock_requests, mock_sleep, mock_logger, error_response):
        mock_requests.side_effect = [Exception("could not attempt request"), error_response, error_response]
        with pytest.raises(requests.HTTPError):
            http.get("http://example.com/some-path", mock_logger, retries=2, backoff_in_seconds=3)
        mock_logger.error.assert_called()

    @patch("time.sleep")
    @patch("requests.get")
    def test_correct_number_of_retria(self, mock_requests, mock_sleep, mock_logger, error_response):
        mock_requests.side_effect = [
            error_response,
            error_response,
            error_response,
            error_response,
            error_response,
        ]  # more than enough
        with pytest.raises(requests.HTTPError):
            http.get("http://example.com/some-path", mock_logger, retries=2, backoff_in_seconds=3)
        assert len(mock_requests.call_args_list) == 3  # once for initial plus two retries

    @patch("time.sleep")
    @patch("requests.get")
    @patch("random.uniform")
    def test_succeeds_if_retries_succeed(
        self, mock_uniform_random, mock_requests, mock_sleep, mock_logger, error_response, success_response
    ):
        mock_uniform_random.side_effect = [0.1]
        mock_requests.side_effect = [error_response, success_response]
        http.get("http://example.com/some-path", mock_logger, retries=1, backoff_in_seconds=22)
        mock_sleep.assert_called_with(22.1)
        mock_logger.warning.assert_called()
        mock_logger.error.assert_not_called()
        mock_requests.assert_called_with("http://example.com/some-path", timeout=http.DEFAULT_TIMEOUT)

    @patch("requests.get")
    def test_timeout_is_passed_in(self, mock_requests, mock_logger):
        http.get("http://example.com/some-path", mock_logger, timeout=12345)
        mock_requests.assert_called_with("http://example.com/some-path", timeout=12345)

    @patch("time.sleep")
    @patch("requests.get")
    @patch("random.uniform")
    def test_exponential_backoff_and_jitter(
        self, mock_uniform_random, mock_requests, mock_sleep, mock_logger, error_response, success_response
    ):
        mock_requests.side_effect = [error_response, error_response, error_response, success_response]
        mock_uniform_random.side_effect = [0.5, 0.4, 0.1]
        http.get("http://example.com/some-path", mock_logger, backoff_in_seconds=10, retries=3)
        assert mock_sleep.call_args_list == [call(10 + 0.5), call(10 * 2 + 0.4), call(10 * 4 + 0.1)]

    @patch("time.sleep")
    @patch("requests.get")
    def test_it_logs_the_url_on_failure(self, mock_requests, mock_sleep, mock_logger, error_response):
        mock_requests.side_effect = [error_response, error_response, error_response]
        url = "http://example.com/some-path"
        with pytest.raises(requests.HTTPError):
            http.get(url, mock_logger, retries=2)

        assert url in mock_logger.error.call_args.args[0]

    @patch("time.sleep")
    @patch("requests.get")
    def test_it_log_warns_errors(self, mock_requests, mock_sleep, mock_logger, error_response, success_response):
        mock_requests.side_effect = [error_response, success_response]
        http.get("http://example.com/some-path", mock_logger, retries=1, backoff_in_seconds=33)

        logged_warnings = [call.args[0] for call in mock_logger.warning.call_args_list]

        assert any("HTTP ERROR" in message for message in logged_warnings), "Expected 'HTTP ERROR' in logged warnings."
        assert any(
            "will retry in 33 seconds" in message for message in logged_warnings
        ), "Expected retry message in logged warnings."

    @patch("time.sleep")
    @patch("requests.get")
    def test_it_calls_status_handler(self, mock_requests, mock_sleep, mock_logger, error_response, success_response):
        mock_requests.side_effect = [error_response]
        status_handler = MagicMock()
        result = http.get(
            "http://example.com/some-path", mock_logger, status_handler=status_handler, retries=1, backoff_in_seconds=33
        )
        mock_sleep.assert_not_called()
        status_handler.assert_called_once()
        assert status_handler.call_args.args[0] == error_response
        assert result == error_response

    @patch("time.sleep")
    @patch("requests.get")
    @patch("random.uniform")
    def test_it_retries_when_status_handler_raises(
        self, mock_uniform_random, mock_requests, mock_sleep, mock_logger, error_response, success_response
    ):
        mock_uniform_random.side_effect = [0.25]
        mock_requests.side_effect = [success_response, error_response]
        status_handler = MagicMock()
        status_handler.side_effect = [Exception("custom exception"), None]
        result = http.get(
            "http://example.com/some-path", mock_logger, status_handler=status_handler, retries=1, backoff_in_seconds=33
        )
        mock_sleep.assert_called_with(33.25)
        # custom status handler raised the first time it was called,
        # so we expect the second mock response to be returned overall
        assert result == error_response


@pytest.mark.parametrize(
    "interval, jitter, max_value, expected",
    [
        (
            30,  # interval
            False,  # jitter
            None,  # max_value
            [30, 60, 120, 240, 480, 960, 1920, 3840, 7680, 15360, 30720, 61440, 122880, 245760, 491520],  # expected
        ),
        (
            3,  # interval
            False,  # jitter
            1000,  # max_value
            [3, 6, 12, 24, 48, 96, 192, 384, 768, 1000, 1000, 1000, 1000, 1000, 1000],  # expected
        ),
    ],
)
def test_backoff_sleep_interval(interval, jitter, max_value, expected):
    actual = [
        http.backoff_sleep_interval(interval, attempt, jitter=jitter, max_value=max_value) for attempt in range(len(expected))
    ]

    if not jitter:
        assert actual == expected
    else:
        for i, (a, e) in enumerate(zip(actual, expected)):
            assert a >= e and a <= e + 1, f"Jittered value out of bounds at attempt {i}: {a} (expected ~{e})"



# ruff: noqa: INP001

import logging
from unittest import mock

import awesome
import orjson

from vunnel import provider, result

fakedata = [
    {
        "name": "FAKE-SA-001",
        "packages": ["curl"],
        "severity": "Critical",
        "description": "Bad thing, really bad thing",
        "affected": "1.0",
        "fixed": "2.0",
    },
    {
        "name": "FAKE-SA-002",
        "packages": ["wget"],
        "severity": "Low",
        "description": "Not really a bad thing, but no fix yet",
        "affected": "5.0",
        "fixed": None,
    },
]


def main():
    # we just want to show the parser working, but not against any real data. For that reason we'll mock
    # this like we would in a unit test.
    with mock.patch("awesome.parser.requests.get") as get:
        get.return_value = mock.Mock(
            json=lambda: fakedata,
            text=orjson.dumps(fakedata),
            raise_for_status=lambda: None,
            status_code=200,
        )

        config = awesome.Config(
            runtime=provider.RuntimeConfig(
                result_store=result.StoreStrategy.FLAT_FILE,
                existing_results=provider.ResultStatePolicy.DELETE_BEFORE_WRITE,
            ),
        )

        my_provider = awesome.Provider(root="./data", config=config)
        my_provider.run()


if __name__ == "__main__":

    logging.basicConfig(
        level=logging.DEBUG,
        format="[%(levelname)s] %(message)s",
        handlers=[logging.StreamHandler()],
    )

    main()



from __future__ import annotations

import os
import tarfile
import tempfile
from pathlib import Path

import zstandard


def extract(path: str, destination_dir: str) -> None:
    if path.endswith(".tar.zst"):
        return _extract_tar_zst(path, destination_dir)

    # open for reading with transparent compression (supports gz, bz2, and xz)
    with tarfile.open(path, mode="r:*") as tar:
        _safe_extract_tar(tar, destination_dir)

    return None


def _extract_tar_zst(path: str, unarchive_path: str) -> None:
    archive_path = Path(path).expanduser()
    dctx = zstandard.ZstdDecompressor(max_window_size=2147483648)

    with tempfile.TemporaryFile(suffix=".tar") as ofh:
        with archive_path.open("rb") as ifh:
            dctx.copy_stream(ifh, ofh)
        ofh.seek(0)
        with tarfile.open(fileobj=ofh, mode="r") as z:
            return _safe_extract_tar(z, unarchive_path)


def _safe_extract_tar(tar: tarfile.TarFile, destination_dir: str) -> None:
    # explanation of noqa: S202
    # This function is a safe wrapper around tar.extractall.
    tar.extractall(destination_dir, filter=_filter_path_traversal)  # noqa: S202


def _filter_path_traversal(tarinfo: tarfile.TarInfo, path: str) -> tarfile.TarInfo | None:
    # drop any path that would result in a write outside the destination dir
    # e.g.
    #   allowed: './some-dir/file.txt'
    #   not allowed: 'some-dir/../../../../../etc/passwd'
    dest_dir = Path(os.path.abspath(path))
    write_path = Path(os.path.normpath(os.path.join(dest_dir, tarinfo.name)))

    if dest_dir in write_path.parents:
        return tarinfo
    return None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from vunnel import provider, result, schema

from .parser import Parser

if TYPE_CHECKING:
    import datetime

# NOTE, CHANGE ME!: a unique and semantically useful name for this provider
PROVIDER_NAME = "my-awesome-provider"

# NOTE, CHANGE ME!: the data shape that all entries produced by this provider conform to
SCHEMA = schema.OSSchema()


@dataclass
class Config:
    runtime: provider.RuntimeConfig = field(
        default_factory=lambda: provider.RuntimeConfig(
            result_store=result.StoreStrategy.SQLITE,
            existing_results=provider.ResultStatePolicy.DELETE_BEFORE_WRITE,
        ),
    )
    request_timeout: int = 125

    # NOTE, CHANGE ME!: Example for fetching secrets from the environment and sanitizing output.
    # It is important to sanitize the __str__ method so that these secrets are not accidentally
    # written to log output.
    #
    # token: str = "env:VUNNEL_AWESOME_TOKEN"
    #
    # def __post_init__(self) -> None:
    #     if self.token.startswith("env:"):
    #         self.token = os.environ.get(self.token[4:], "")
    #
    # def __str__(self) -> str:
    #     # sanitize secrets from any output
    #     tok_value = self.token
    #     str_value = super().__str__()
    #     if not tok_value:
    #         return str_value
    #     return str_value.replace(tok_value, "********")


class Provider(provider.Provider):
    def __init__(self, root: str, config: Config | None = None):
        if not config:
            config = Config()

        super().__init__(root, runtime_cfg=config.runtime)
        self.config = config
        self.logger.debug(f"config: {config}")

        self.parser = Parser(
            ws=self.workspace,
            download_timeout=self.config.request_timeout,
            logger=self.logger,
            # NOTE, CHANGE ME!: example of passing a config secret to the parser to download the vulnerability data
            # token=self.config.token
        )

        # this provider requires the previous state from former runs
        provider.disallow_existing_input_policy(config.runtime)

    @classmethod
    def name(cls) -> str:
        return PROVIDER_NAME

    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:

        # NOTE: CHANGE ME! Why is last_updated passed in here? This allows you to be able to make decisions about
        # incremental updates of the existing vulnerability data state instead of needing to download all
        # vulnerability data from the source. For an example of this see the NVD provider implementation at
        # https://github.com/anchore/vunnel/blob/main/src/vunnel/providers/nvd/manager.py

        with self.results_writer() as writer:

            for vuln_id, record in self.parser.get():
                vuln_id = vuln_id.lower()

                writer.write(
                    identifier=vuln_id,
                    schema=SCHEMA,
                    payload=record,
                )

        return self.parser.urls, len(writer)



from __future__ import annotations

import os
import json

import pytest
from vunnel import provider, result
from vunnel.providers import nvd


@pytest.mark.parametrize(
    ("policy", "should_raise"),
    (
        (provider.ResultStatePolicy.KEEP, False),
        (provider.ResultStatePolicy.DELETE_BEFORE_WRITE, True),
        (provider.ResultStatePolicy.DELETE, True),
    ),
)
def test_incremental_update_with_existing_results(policy, should_raise):
    def make():
        nvd.Provider(
            "/tmp/doesntmatter",
            nvd.Config(runtime=provider.RuntimeConfig(existing_results=policy, result_store=result.StoreStrategy.SQLITE)),
        )

    if should_raise:
        with pytest.raises(Exception):
            make()
    else:
        make()


@pytest.mark.parametrize(
    ("store", "should_raise"),
    (
        (result.StoreStrategy.FLAT_FILE, True),
        (result.StoreStrategy.SQLITE, False),
    ),
)
def test_require_sqlite_store(store, should_raise):
    def make():
        nvd.Provider("/tmp/doesntmatter", nvd.Config(runtime=provider.RuntimeConfig(result_store=store)))

    if should_raise:
        with pytest.raises(Exception):
            make()
    else:
        make()


@pytest.mark.parametrize(
    ("overrides_enabled", "overrides_url", "should_raise"),
    (
        (True, "something", False),
        (False, "something", False),
        (True, "", True),
        (False, "", False),
    ),
)
def test_require_override_configuration(overrides_enabled, overrides_url, should_raise):
    def make():
        nvd.Provider(
            "/tmp/doesntmatter",
            nvd.Config(
                overrides_enabled=overrides_enabled,
                overrides_url=overrides_url,
                runtime=provider.RuntimeConfig(result_store=result.StoreStrategy.SQLITE),
            ),
        )

    if should_raise:
        with pytest.raises(Exception):
            make()
    else:
        make()


@pytest.mark.parametrize(
    ("mock_data_path", "expected_written_entries"),
    [
        ("test-fixtures/single-entry.json", 1),
    ],
)
def test_provider_schema(helpers, mock_data_path, expected_written_entries, disable_get_requests, mocker):
    workspace = helpers.provider_workspace_helper(name=nvd.Provider.name())
    mock_data_path = helpers.local_dir(mock_data_path)

    with open(mock_data_path) as f:
        json_dict = json.load(f)

    c = nvd.Config()
    p = nvd.Provider(root=workspace.root, config=c)
    p.config.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p.manager.api.cve = mocker.Mock(return_value=[json_dict])

    p.update(None)

    assert expected_written_entries == workspace.num_result_entries()
    assert workspace.result_schemas_valid(require_entries=expected_written_entries > 0)


@pytest.mark.parametrize(
    ("mock_data_path",),
    [
        ("test-fixtures/single-entry.json",),
        ("test-fixtures/full-page.json",),
    ],
)
def test_provider_via_snapshot(helpers, mock_data_path, disable_get_requests, mocker):
    snapshot_name = os.path.basename(mock_data_path).split(".")[0]

    workspace = helpers.provider_workspace_helper(
        name=nvd.Provider.name(),
        snapshot_prefix=snapshot_name,
    )

    c = nvd.Config()
    p = nvd.Provider(
        root=workspace.root,
        config=c,
    )
    # keep all of the default values for the result store, but override the strategy
    p.config.runtime.result_store = result.StoreStrategy.FLAT_FILE

    mock_data_path = helpers.local_dir(mock_data_path)

    with open(mock_data_path) as f:
        json_dict = json.load(f)

    p.manager.api.cve = mocker.Mock(return_value=[json_dict])

    p.update(None)

    workspace.assert_result_snapshots()



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from vunnel import provider, result, schema

from .parser import Parser

if TYPE_CHECKING:
    import datetime


@dataclass
class Config:
    runtime: provider.RuntimeConfig = field(
        default_factory=lambda: provider.RuntimeConfig(
            result_store=result.StoreStrategy.SQLITE,
            existing_results=result.ResultStatePolicy.DELETE_BEFORE_WRITE,
        ),
    )
    request_timeout: int = 125


class Provider(provider.Provider):
    __schema__ = schema.OSVSchema()
    __distribution_version__ = int(__schema__.major_version)

    def __init__(self, root: str, config: Config | None = None):
        if not config:
            config = Config()

        super().__init__(root, runtime_cfg=config.runtime)
        self.config = config
        self.logger.debug(f"config: {config}")

        self.schema = self.__schema__
        self.parser = Parser(
            ws=self.workspace,
            logger=self.logger,
        )

        # this provider requires the previous state from former runs
        provider.disallow_existing_input_policy(config.runtime)

    @classmethod
    def name(cls) -> str:
        return "alma"

    @classmethod
    def compatible_schema(cls, schema_version: str) -> schema.Schema | None:
        candidate = schema.OSVSchema(schema_version)
        if candidate.major_version == cls.__schema__.major_version:
            return candidate
        return None

    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:
        # TODO: use of last_updated as NVD provider does to avoid downloading all
        # vulnerability data from the source and make incremental updates instead
        with self.results_writer() as writer:
            for vuln_id, vuln_schema_version, record in self.parser.get():
                vuln_schema = self.compatible_schema(vuln_schema_version)
                if not vuln_schema:
                    self.logger.warning(
                        f"skipping vulnerability {vuln_id} with schema version {vuln_schema_version} ",
                        f"as is incompatible with provider schema version {self.schema.version}",
                    )
                    continue
                writer.write(
                    identifier=vuln_id.lower(),
                    schema=vuln_schema,
                    payload=record,
                )

        return self.parser.urls, len(writer)



from __future__ import annotations

import datetime
import os

import pytest

from vunnel import result, schema, workspace


def assert_directory(path: str, exists: bool = True, empty: bool = False):
    if not exists and not empty:
        raise ValueError("cannot assert directory does not exist and is not empty")

    if exists:
        assert os.path.exists(path), f"directory does not exist (but should): {path}"
        if empty:
            assert len(os.listdir(path)) == 0, f"directory is not empty (but should be): {path}"
        else:
            assert len(os.listdir(path)) > 0, f"directory is empty (but should not be): {path}"
    else:
        assert not os.path.exists(path), f"directory exists (but should not): {path}"


def test_clear_input(tmpdir, dummy_file):
    ws = workspace.Workspace(root=tmpdir, name="dummy", create=True)

    # create a dummy files
    dummy_file(ws.input_path)
    dummy_file(ws.results_path)

    assert_directory(ws.input_path, exists=True, empty=False)
    assert_directory(ws.results_path, exists=True, empty=False)

    # clear the input
    ws.clear_input()

    assert_directory(ws.input_path, exists=True, empty=True)
    assert_directory(ws.results_path, exists=True, empty=False)


def test_clear_results(tmpdir, dummy_file):
    ws = workspace.Workspace(root=tmpdir, name="dummy", create=True)

    # create a dummy files
    dummy_file(ws.input_path)
    dummy_file(ws.results_path)

    urls = ["http://localhost:8000/dummy-input-1.json"]
    store = result.StoreStrategy.FLAT_FILE
    ws.record_state(urls=urls, store=store.value, timestamp=datetime.datetime(2021, 1, 1), version=1, distribution_version=1)

    assert_directory(ws.input_path, exists=True, empty=False)
    assert_directory(ws.results_path, exists=True, empty=False)

    # clear the results
    ws.clear_results()

    assert_directory(ws.input_path, exists=True, empty=False)
    assert_directory(ws.results_path, exists=True, empty=True)

    # ensure the URLs are still populated
    new_state = workspace.State.read(root=ws.path)
    assert urls == new_state.urls


def test_record_state(tmpdir, dummy_file):
    ws = workspace.Workspace(root=tmpdir, name="dummy", create=True)

    # create a dummy files
    dummy_file(ws.input_path, "dummt-input-1.json")
    dummy_file(ws.results_path, "dummy-00000.json")

    urls = ["http://localhost:8000/dummy-input-1.json"]
    store = result.StoreStrategy.FLAT_FILE
    ws.record_state(urls=urls, store=store.value, timestamp=datetime.datetime(2021, 1, 1), version=1, distribution_version=1)

    current_state = workspace.State.read(root=ws.path)

    # ignore timestamp (make certain it exists)
    assert current_state.timestamp is not None
    current_state.timestamp = None

    expected_state = workspace.State(
        store=result.StoreStrategy.FLAT_FILE.value,
        provider="dummy",
        urls=["http://localhost:8000/dummy-input-1.json"],
        listing=workspace.File(digest="63b7adef165e430a", algorithm="xxh64", path="checksums"),
        timestamp=None,
        schema=schema.ProviderStateSchema(),
    )

    assert current_state == expected_state


def test_record_state_urls_persisted_across_runs(tmpdir, dummy_file):
    ws = workspace.Workspace(root=tmpdir, name="dummy", create=True)

    # create a dummy files
    dummy_file(ws.input_path, "dummt-input-1.json")
    dummy_file(ws.results_path, "dummy-00000.json")

    urls = ["http://localhost:8000/dummy-input-1.json"]
    store = result.StoreStrategy.FLAT_FILE
    ws.record_state(urls=urls, store=store.value, timestamp=datetime.datetime(2021, 1, 1), version=1, distribution_version=1)

    # this call should not clear the URLs
    ws.record_state(urls=None, store=store.value, timestamp=datetime.datetime(2021, 1, 1), version=1, distribution_version=1)

    current_state = workspace.State.read(root=ws.path)

    # ignore timestamp (make certain it exists)
    assert current_state.timestamp is not None
    current_state.timestamp = None

    expected_state = workspace.State(
        store=store,
        provider="dummy",
        urls=["http://localhost:8000/dummy-input-1.json"],
        listing=workspace.File(digest="63b7adef165e430a", algorithm="xxh64", path="checksums"),
        timestamp=None,
        schema=schema.ProviderStateSchema(),
    )

    assert current_state == expected_state


def test_state_schema(tmpdir, dummy_file, helpers):
    name = "dummy"
    ws = workspace.Workspace(root=tmpdir, name=name, create=True)

    # create a dummy files
    dummy_file(ws.input_path, "dummt-input-1.json")
    dummy_file(ws.results_path, "dummy-00000.json")

    urls = ["http://localhost:8000/dummy-input-1.json"]
    store = result.StoreStrategy.FLAT_FILE
    ws.record_state(urls=urls, store=store.value, timestamp=datetime.datetime(2021, 1, 1), version=1, distribution_version=1)

    ws_helper = helpers.provider_workspace_helper(name=name, create=False)

    assert ws_helper.metadata_schema_valid()


def test_replace_results(tmpdir, dummy_file):
    # make 2 workspaces
    ws1_dir = tmpdir.mkdir("ws1")
    ws1 = workspace.Workspace(root=ws1_dir, name="old", create=True)
    dummy_file(ws1.input_path, "old-input-1.json")
    dummy_file(ws1.results_path, "old-00000.json")
    urls = ["http://localhost:8000/old_workspace.json"]
    store = result.StoreStrategy.FLAT_FILE
    ws1.record_state(urls=urls, store=store.value, timestamp=datetime.datetime(2021, 1, 1), version=1, distribution_version=1)

    ws2_dir = tmpdir.mkdir("ws2")
    ws2 = workspace.Workspace(root=ws2_dir, name="new", create=True)
    dummy_file(ws2.input_path, "new-input-2.json")
    urls = ["http://localhost:8000/new_workspace.json"]
    store = result.StoreStrategy.FLAT_FILE
    dummy_file(ws2.results_path, "new-00000.json")
    ws2.record_state(urls=urls, store=store.value, timestamp=datetime.datetime(2024, 1, 1), version=1, distribution_version=1)
    # be able to tell the contents apart
    # replace the results in one with the other
    # assert on the contents
    ws1.replace_results(ws2)
    # assert that metadata / state is replaced
    assert "http://localhost:8000/new_workspace.json" in ws1.state().urls
    assert "http://localhost:8000/old_workspace.json" not in ws1.state().urls

    # assert that results have been replaced
    assert os.path.exists(os.path.join(ws1.results_path, "new-00000.json"))
    assert not os.path.exists(os.path.join(ws1.results_path, "old-00000.json"))

    # assert that input path is not changed
    assert os.path.exists(os.path.join(ws1.input_path, "old-input-1.json"))
    assert not os.path.exists(os.path.join(ws1.input_path, "new-input-1.json"))


def test_validate_checksums_valid(tmpdir, dummy_file):
    ws = workspace.Workspace(root=tmpdir, name="dummy", create=True)
    urls = []
    dummy_file(ws.results_path, "dummy-result-1.json")
    ws.record_state(
        urls=urls,
        store=result.StoreStrategy.FLAT_FILE.value,
        timestamp=datetime.datetime(2021, 1, 1),
        version=1,
        distribution_version=1,
    )
    ws.validate_checksums()


def test_validate_checksums_invalid(tmpdir, dummy_file):
    ws = workspace.Workspace(root=tmpdir, name="dummy", create=True)
    urls = []
    result_file_path = dummy_file(ws.results_path, "dummy-result-1.json")
    ws.record_state(
        urls=urls,
        store=result.StoreStrategy.FLAT_FILE.value,
        timestamp=datetime.datetime(2021, 1, 1),
        version=1,
        distribution_version=1,
    )
    with open(result_file_path, "a") as f:
        f.write("invalid")
    with pytest.raises(RuntimeError) as e:
        ws.validate_checksums()
        assert "has been modified" in str(e)


def test_validate_checksums_missing(tmpdir, dummy_file):
    ws = workspace.Workspace(root=tmpdir, name="dummy", create=True)
    urls = []
    result_file_path = dummy_file(ws.results_path, "dummy-result-1.json")
    ws.record_state(
        urls=urls,
        store=result.StoreStrategy.FLAT_FILE.value,
        timestamp=datetime.datetime(2021, 1, 1),
        version=1,
        distribution_version=1,
    )
    os.remove(result_file_path)
    with pytest.raises(RuntimeError) as e:
        ws.validate_checksums()
        assert "does not exist" in str(e)


def test_validate_checksums_input_changed_ok(tmpdir, dummy_file):
    ws = workspace.Workspace(root=tmpdir, name="dummy", create=True)
    urls = []
    result_file_path = dummy_file(ws.results_path, "dummy-result-1.json")
    input_file_path = dummy_file(ws.input_path, "dummy-input-1.json")
    ws.record_state(
        urls=urls,
        store=result.StoreStrategy.FLAT_FILE.value,
        timestamp=datetime.datetime(2021, 1, 1),
        version=1,
        distribution_version=1,
    )
    with open(input_file_path, "a") as f:
        f.write("this won't change the checksums")

    ws.validate_checksums()



from __future__ import annotations

import os
import shutil

import pytest
from vunnel import result, workspace
from vunnel.providers.alpine import Config, Provider, parser
from vunnel.providers.alpine.parser import Parser, SecdbLandingParser


class TestAlpineProvider:
    @pytest.fixture()
    def mock_raw_data(self):
        """
        Returns stringified version of the following yaml

        ---
        distroversion: v0.0
        reponame: main
        archs:
          - x86_64
          - x86
          - armhf
        urlprefix: http://dl-cdn.alpinelinux.org/alpine
        apkurl: "{{urlprefix}}/{{distroversion}}/{{reponame}}/{{arch}}/{{pkg.name}}-{{pkg.ver}}.apk"
        packages:
          - pkg:
              name: apache2
              secfixes:
                2.4.26-r0:
                  - CVE-2017-3167
                  - CVE-2017-3169
                  - CVE-2017-7659
                  - CVE-2017-7668
                  - CVE-2017-7679
                2.4.27-r0:
                  - CVE-2017-9789
                2.4.27-r1:
                  - CVE-2017-9798
          - pkg:
              name: augeas
              secfixes:
                1.4.0-r5:
                - CVE-2017-7555
          - pkg:
              name: bash
              secfixes:
                4.3.42-r5:
                  - CVE-2016-9401
        """

        return "apkurl: '{{urlprefix}}/{{distroversion}}/{{reponame}}/{{arch}}/{{pkg.name}}-{{pkg.ver}}.apk'\narchs:\n- x86_64\n- x86\n- armhf\ndistroversion: v0.0\npackages:\n- pkg:\n    name: apache2\n    secfixes:\n      2.4.26-r0:\n      - CVE-2017-3167\n      - CVE-2017-3169\n      - CVE-2017-7659\n      - CVE-2017-7668\n      - CVE-2017-7679\n      2.4.27-r0:\n      - CVE-2017-9789\n      2.4.27-r1:\n      - CVE-2017-9798\n- pkg:\n    name: augeas\n    secfixes:\n      1.4.0-r5:\n      - CVE-2017-7555\n- pkg:\n    name: bash\n    secfixes:\n      4.3.42-r5:\n      - CVE-2016-9401\nreponame: main\nurlprefix: http://dl-cdn.alpinelinux.org/alpine\n"  # noqa: E501

    @pytest.fixture()
    def mock_parsed_data(self):
        """
        Returns the parsed output generated by AlpineDataProvider._load() for the mock_raw_data

        :return:
        """
        release = "0.0"
        dbtype_data_dict = {
            "main": {
                "apkurl": "{{urlprefix}}/{{distroversion}}/{{reponame}}/{{arch}}/{{pkg.name}}-{{pkg.ver}}.apk",
                "archs": ["x86_64", "x86", "armhf"],
                "distroversion": "v0.0",
                "packages": [
                    {
                        "pkg": {
                            "name": "apache2",
                            "secfixes": {
                                "2.4.26-r0": [
                                    "CVE-2017-3167",
                                    "CVE-2017-3169",
                                    "CVE-2017-7659",
                                    "CVE-2017-7668",
                                    "CVE-2017-7679",
                                ],
                                "2.4.27-r0": ["CVE-2017-9789"],
                                "2.4.27-r1": ["CVE-2017-9798"],
                            },
                        },
                    },
                    {
                        "pkg": {
                            "name": "augeas",
                            "secfixes": {"1.4.0-r5": ["CVE-2017-7555"]},
                        },
                    },
                    {
                        "pkg": {
                            "name": "bash",
                            "secfixes": {"4.3.42-r5": ["CVE-2016-9401"]},
                        },
                    },
                ],
                "reponame": "main",
                "urlprefix": "http://dl-cdn.alpinelinux.org/alpine",
            },
        }
        return release, dbtype_data_dict

    @pytest.mark.parametrize(
        ("release", "expected"),
        [
            ("v3.3", True),
            ("3.4", False),
            ("v.3.3", False),
            ("v3.101", True),
            ("v3.27", True),
        ],
    )
    def test_release_regex(self, release, expected):
        assert bool(Parser._release_regex_.match(release)) == expected

    def test_load(self, mock_raw_data, tmpdir):
        p = Parser(workspace=workspace.Workspace(tmpdir, "test", create=True))
        a = os.path.join(p.secdb_dir_path, "v0.0")
        os.makedirs(a, exist_ok=True)
        b = os.path.join(a, "main.yaml")
        with open(b, "w") as fp:
            fp.write(mock_raw_data)

        counter = 0
        for release, dbtype_data_dict in p._load():
            counter += 1
            print(f"got secdb data for release {release}, db types: {list(dbtype_data_dict.keys())}")
            assert release == "0.0"
            assert isinstance(dbtype_data_dict, dict)
            assert list(dbtype_data_dict.keys()) == ["main"]
            assert all("packages" in x for x in dbtype_data_dict.values())

        assert counter == 1

    def test_normalize(self, mock_parsed_data, tmpdir):
        p = Parser(workspace=workspace.Workspace(tmpdir, "test", create=True))
        release = mock_parsed_data[0]
        dbtype_data_dict = mock_parsed_data[1]

        vuln_records = p._normalize(release, dbtype_data_dict)
        assert len(vuln_records) > 0
        assert all("Vulnerability" in x for x in vuln_records.values())
        assert sorted(vuln_records.keys()) == sorted(
            [
                "CVE-2017-3167",
                "CVE-2017-3169",
                "CVE-2017-7659",
                "CVE-2017-7668",
                "CVE-2017-7679",
                "CVE-2017-9789",
                "CVE-2017-9798",
                "CVE-2017-7555",
                "CVE-2016-9401",
            ],
        )

    @pytest.mark.parametrize(
        ("content", "expected"),
        [
            pytest.param(
                '<html>\r\n<head><title>Index of /</title></head>\r\n<body>\r\n<h1>Index of /</h1><hr><pre><a href="../">../</a>\r\n<a href="v3.10/">v3.10/</a> 11-Jun-2020 20:17 -\r\n<a href="v3.11/">v3.11/</a> 11-Jun-2020 18:12 -\r\n</pre><hr></body>\r\n</html>\r\n',  # noqa: E501
                ["v3.10/", "v3.11/"],
                id="with-content",
            ),
            pytest.param('<a href=".">.</a>', [], id="href-."),
            pytest.param('<a href="../">../</a>', [], id="href-../"),
            pytest.param('<a href="foo/">foo/</a>', ["foo/"], id="href-foo/"),
        ],
    )
    def test_secdb_landing_parser(self, content, expected):
        parser = SecdbLandingParser()
        parser.feed(content)
        assert sorted(parser.links) == sorted(expected)

    @pytest.mark.parametrize(
        ("content", "expected"),
        [
            pytest.param(
                '<html>\r\n<head><title>Index of /</title></head>\r\n<body>\r\n<h1>Index of /</h1><hr><pre><a href="../">../</a>\r\n<a href="v3.10/">v3.10/</a> 11-Jun-2020 20:17 -\r\n<a href="v3.11/">v3.11/</a> 11-Jun-2020 18:12 -\r\n</pre><hr></body>\r\n</html>\r\n',  # noqa: E501
                ["v3.10/", "v3.11/"],
                id="with-content",
            ),
            pytest.param('<a href=".">.</a>', [], id="href-."),
            pytest.param('<a href=    "../"   >../</a>', [], id="href-../+space"),
            pytest.param('<a href=  "foo/" >foo/</a>', ["foo/"], id="href-foo/+space"),
        ],
    )
    def test_link_finder_regex(self, content, expected):
        assert Parser._link_finder_regex_.findall(content) == expected


def test_provider_schema(helpers, disable_get_requests, monkeypatch):
    workspace = helpers.provider_workspace_helper(name=Provider.name())

    c = Config()
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(
        root=workspace.root,
        config=c,
    )

    mock_data_path = helpers.local_dir("test-fixtures/input")
    shutil.copytree(mock_data_path, workspace.input_dir, dirs_exist_ok=True)

    def mock_download():
        return

    monkeypatch.setattr(p.parser, "_download", mock_download)

    p.update(None)

    assert workspace.num_result_entries() == 16
    assert workspace.result_schemas_valid(require_entries=True)


def test_provider_via_snapshot(helpers, disable_get_requests, monkeypatch):
    workspace = helpers.provider_workspace_helper(
        name=Provider.name(),
        input_fixture="test-fixtures/input",
    )

    c = Config()
    # keep all of the default values for the result store, but override the strategy
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(
        root=workspace.root,
        config=c,
    )

    def mock_download():
        return None

    monkeypatch.setattr(p.parser, "_download", mock_download)
    monkeypatch.setattr(p.parser, "_download_metadata_url", mock_download)
    monkeypatch.setattr(p.parser, "_download_url", mock_download)

    p.update(None)

    workspace.assert_result_snapshots()



import os

import pytest
from pathlib import Path
from unittest.mock import MagicMock
from vunnel.providers.rhel.csaf_client import CSAFClient

@pytest.fixture()
def fixture_dir():
    return Path(__file__).parent / "test-fixtures"

@pytest.fixture()
def mock_workspace(helpers, fixture_dir):
    input_path = fixture_dir / "csaf" / "input"
    os.makedirs(input_path, exist_ok=True)

    return helpers.provider_workspace_helper(
            name="rhel",
            input_fixture=input_path,
        )

@pytest.fixture
def mock_http_get(mocker, fixture_dir):
    mock = mocker.patch('vunnel.utils.http_wrapper.get')

    fs = fixture_dir / "csaf/server"

    def request_side_effect(url, *args, **kwargs):
        requested_file = fs / url.removeprefix("https://example.com/advisories/")
        if not requested_file.exists():
            return MagicMock(status_code=404)
        if kwargs.get('stream'):
            response_mock = MagicMock(status_code=200)
            response_mock.iter_content = lambda chunk_size=65536: (
                requested_file.read_bytes()[i:i+chunk_size] for i in range(0, len(requested_file.read_bytes()), chunk_size)
            )
            response_mock.__enter__.return_value = response_mock # support "with"
            return response_mock
        return MagicMock(status_code=200, text=requested_file.read_text())

    mock.side_effect = request_side_effect
    return mock

@pytest.fixture
def mock_os_remove(mocker):
    return mocker.patch('os.remove')

@pytest.fixture
def mock_tar_extract(mocker):
    return mocker.patch('vunnel.providers.rhel.csaf_client.extract', autospec=True)

@pytest.fixture()
def latest_file_name(fixture_dir):
    return (fixture_dir / "csaf/server/archive_latest.txt").read_text()

def test_process_changes_and_deletions(mock_workspace, mock_http_get, mock_os_remove, mock_tar_extract, latest_file_name):
    latest_url = "https://example.com/advisories/archive_latest.txt"
    expected_tar_url = latest_url.replace(latest_file_name, "archive_latest.txt")
    changes_url = latest_url.replace("archive_latest.txt", "changes.csv")
    deletions_url = latest_url.replace("archive_latest.txt", "deletions.csv")
    changed_rhsa_url = latest_url.replace("archive_latest.txt", "2024/rhsa-2024_11255.json")
    other_changed_rhsa_url = latest_url.replace("archive_latest.txt", "2024/rhsa-2024_0010.json")
    deleted_rhsa_path = mock_workspace.input_path / "advisories" / "2024/rhsa-2024_2106.json"

    client = CSAFClient(workspace=mock_workspace, logger=MagicMock(), latest_url=latest_url)
    # Extract only the first argument (URL) from all calls
    called_urls = [args[0] for args, _ in mock_http_get.call_args_list]

    # Assert the expected URLs were requested
    assert latest_url in called_urls
    assert expected_tar_url in called_urls
    assert changes_url in called_urls
    assert deletions_url in called_urls
    # assert both changed files get re-downloaded
    assert changed_rhsa_url in called_urls
    assert other_changed_rhsa_url in called_urls

    # assert that it stops processing rows of changes.csv after encountering a data
    # before the archive latest date
    too_old_entries = [
        "2024/rhsa-2024_4444.json",
        "2024/rhsa-2024_5555.json",
        "2024/rhsa-2024_6666.json",
    ]
    too_old_changes_urls = [latest_url.replace("archive_latest.txt", entry) for entry in too_old_entries]
    assert not any(url in called_urls for url in too_old_changes_urls)

    # Check that os.remove was called for the deleted advisory
    called_paths = [args[0] for args, _ in mock_os_remove.call_args_list]
    assert deleted_rhsa_path in called_paths

    # assert that we can fetch the correct RHSA by ID
    doc = client.csaf_doc_for_rhsa("RHSA-2024:0010")
    assert doc.document.aggregate_severity.text == "Important"
    doc_self_link = next((r.url for r in doc.document.references if r.category == "self"), None)
    assert doc_self_link.endswith("RHSA-2024:0010")



import json

import pytest
from datetime import datetime, timezone
from src.vunnel.distribution import ListingEntry, ListingDocument


class TestListingEntry:

    @pytest.fixture
    def entry1(self):
        return ListingEntry(
            built="2022-01-01T00:00:00Z",
            distribution_version=1,
            url="http://example.com/archive.tar.gz",
            distribution_checksum="sha256:1234567890abcdef1234567890abcdef",
            enclosed_checksum="xxhash64:1234567890abcdef",
        )

    @pytest.fixture
    def entry2(self):
        return ListingEntry(
            built="2022-01-01T00:00:00Z",
            distribution_version=1,
            url="http://example.com/archive.tar.zst",
            distribution_checksum="sha256:abcdef1234567890abcdef1234567890",
            enclosed_checksum="xxhash64:abcdef1234567890",
        )

    def test_basename(self, entry1, entry2):
        assert entry1.basename() == "archive.tar.gz"
        assert entry2.basename() == "archive.tar.zst"

    def test_basename_invalid_url(self, entry1):
        with pytest.raises(RuntimeError):
            entry1.url = "http://example.com/archive.tar.unsupported"
            entry1.basename()

    def test_age_in_days(self, entry1):
        now = datetime.now(tz=timezone.utc)
        assert entry1.age_in_days(now) == (now - datetime(2022, 1, 1, tzinfo=timezone.utc)).days

    def test_age_in_days_no_now(self, entry1):
        assert entry1.age_in_days() == (datetime.now(tz=timezone.utc) - datetime(2022, 1, 1, tzinfo=timezone.utc)).days


class TestListingDocument:
    @pytest.fixture
    def document(self):
        entries = [
            ListingEntry(
                built="2022-01-01T00:00:00Z",
                distribution_version=1,
                url="http://example.com/archive1.tar.gz",
                distribution_checksum="sha256:1234567890abcdef1234567890abcdef",
                enclosed_checksum="xxhash64:1234567890abcdef",
            ),
            ListingEntry(
                built="2022-01-02T00:00:00Z",
                distribution_version=1,
                url="http://example.com/archive2.tar.gz",
                distribution_checksum="sha256:abcdef1234567890abcdef1234567890",
                enclosed_checksum="xxhash64:abcdef1234567890",
            ),
        ]
        return ListingDocument(available={1: entries}, provider="test_provider")

    @pytest.fixture
    def built_document(self):
        subject = ListingDocument.new(provider="nvd")

        subject.add(
            ListingEntry(
                built=datetime(2017, 11, 28, 23, 55, 59, 342380).strftime("%Y-%m-%dT%H:%M:%S.%f%z"),
                distribution_version=3,
                url="https://b-place.com/something-1.tar.gz",
                distribution_checksum="sha256:123456789",
                enclosed_checksum="xxh64:123456789",
            )
        )

        subject.add(
            ListingEntry(
                built=datetime(2016, 11, 28, 23, 55, 59, 342380).strftime("%Y-%m-%dT%H:%M:%S.%f%z"),
                distribution_version=3,
                url="https://a-place.com/something.tar.gz",
                distribution_checksum="sha256:123456789",
                enclosed_checksum="xxh64:123456789",
            )
        )

        subject.add(
            ListingEntry(
                built=datetime(2019, 11, 28, 23, 55, 59, 342380).strftime("%Y-%m-%dT%H:%M:%S.%f%z"),
                distribution_version=3,
                url="https://c-place.com/something.tar.gz",
                distribution_checksum="sha256:123456789",
                enclosed_checksum="xxh64:123456789",
            )
        )

        subject.add(
            ListingEntry(
                built=datetime(2017, 11, 28, 23, 55, 59, 342380).strftime("%Y-%m-%dT%H:%M:%S.%f%z"),
                distribution_version=4,
                url="https://b-place.com/something-1.tar.zst",
                distribution_checksum="sha256:123456789",
                enclosed_checksum="xxh64:123456789",
            )
        )

        subject.add(
            ListingEntry(
                built=datetime(2016, 11, 28, 23, 55, 59, 342380).strftime("%Y-%m-%dT%H:%M:%S.%f%z"),
                distribution_version=4,
                url="https://a-place.com/something.tar.zst",
                distribution_checksum="sha256:123456789",
                enclosed_checksum="xxh64:123456789",
            )
        )

        subject.add(
            ListingEntry(
                built=datetime(2019, 11, 28, 23, 55, 59, 342380).strftime("%Y-%m-%dT%H:%M:%S.%f%z"),
                distribution_version=4,
                url="https://c-place.com/something.tar.zst",
                distribution_checksum="sha256:123456789",
                enclosed_checksum="xxh64:123456789",
            )
        )

        return subject

    def test_latest_entry(self, document, built_document):
        latest_entry = document.latest_entry(1)
        assert latest_entry is not None
        assert latest_entry.distribution_version == 1
        assert latest_entry.url == "http://example.com/archive1.tar.gz"

        assert "https://c-place.com/something.tar.gz" == built_document.latest_entry(3).url
        assert "https://c-place.com/something.tar.zst" == built_document.latest_entry(4).url

    def test_latest_entry_no_entries(self, document):
        latest_entry = document.latest_entry(2)
        assert latest_entry is None

    def test_schema(self, built_document, validate_json_schema):
        content = json.dumps(built_document.to_dict())
        validate_json_schema(content)



import os
import tempfile
import shutil
from unittest.mock import Mock, patch, call

import pytest

from vunnel.providers.rhel import parser


class TestHandleHydraData:

    @pytest.fixture
    def mock_logger(self):
        return Mock()

    @pytest.fixture
    def temp_dirs(self):
        with tempfile.TemporaryDirectory() as temp_root:
            dir_path = os.path.join(temp_root, "main_dir")
            backup_dir_path = os.path.join(temp_root, "backup_dir")
            yield dir_path, backup_dir_path

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    @patch('vunnel.providers.rhel.parser.utils.move_dir')
    def test_ignore_hydra_errors_true_with_existing_dir(self, mock_move_dir, mock_silent_remove,
                                                        mock_exists, mock_makedirs, mock_logger):
        """Test behavior when ignore_hydra_errors=True and directory exists."""
        mock_exists.return_value = True

        dir_path = "/tmp/test/dir"
        backup_dir_path = "/tmp/test/backup"

        with parser.handle_hydra_data(True, dir_path, backup_dir_path, mock_logger):
            pass

        mock_silent_remove.assert_called_with(backup_dir_path, tree=True)
        mock_makedirs.assert_has_calls([call(dir_path)])
        mock_move_dir.assert_called_once_with(dir_path, backup_dir_path)
        mock_logger.debug.assert_called_with(f"moving existing {dir_path} to {backup_dir_path}")

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    @patch('vunnel.providers.rhel.parser.utils.move_dir')
    def test_ignore_hydra_errors_true_without_existing_dir(self, mock_move_dir, mock_silent_remove,
                                                           mock_exists, mock_makedirs, mock_logger):
        """Test behavior when ignore_hydra_errors=True and directory doesn't exist."""
        mock_exists.return_value = False

        dir_path = "/tmp/test/dir"
        backup_dir_path = "/tmp/test/backup"

        with parser.handle_hydra_data(True, dir_path, backup_dir_path, mock_logger):
            pass

        mock_silent_remove.assert_called_once_with(backup_dir_path, tree=True)
        mock_makedirs.assert_has_calls([call(dir_path)])
        mock_move_dir.assert_not_called()
        mock_logger.debug.assert_called_with(f"no existing {dir_path} to move, starting fresh")

    @patch('os.makedirs')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    def test_ignore_hydra_errors_false(self, mock_silent_remove, mock_makedirs, mock_logger):
        """Test behavior when ignore_hydra_errors=False."""
        dir_path = "/tmp/test/dir"
        backup_dir_path = "/tmp/test/backup"

        with parser.handle_hydra_data(False, dir_path, backup_dir_path, mock_logger):
            pass

        # Verify only dir_path is removed and recreated
        mock_silent_remove.assert_called_once_with(dir_path, tree=True)
        mock_makedirs.assert_called_once_with(dir_path)

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    @patch('vunnel.providers.rhel.parser.utils.move_dir')
    def test_exception_with_ignore_hydra_errors_true(self, mock_move_dir, mock_silent_remove,
                                                     mock_exists, mock_makedirs, mock_logger):
        """Test exception handling when ignore_hydra_errors=True."""
        mock_exists.return_value = True

        dir_path = "/tmp/test/dir"
        backup_dir_path = "/tmp/test/backup"
        test_exception = ValueError("Test error")

        # Exception should be caught and handled, not re-raised
        with parser.handle_hydra_data(True, dir_path, backup_dir_path, mock_logger):
            raise test_exception

        # verify restoration calls (2 silent_remove calls: setup + restoration)
        assert mock_silent_remove.call_count == 2
        mock_silent_remove.assert_has_calls([
            call(backup_dir_path, tree=True),  # Setup
            call(dir_path, tree=True)          # Restoration
        ])

        # verify move operations (2 calls: setup + restoration)
        assert mock_move_dir.call_count == 2
        mock_move_dir.assert_has_calls([
            call(dir_path, backup_dir_path),   # Setup
            call(backup_dir_path, dir_path)    # Restoration
        ])

    @patch('os.makedirs')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    def test_exception_with_ignore_hydra_errors_false(self, mock_silent_remove, mock_makedirs, mock_logger):
        """Test exception handling when ignore_hydra_errors=False."""
        dir_path = "/tmp/test/dir"
        backup_dir_path = "/tmp/test/backup"
        test_exception = ValueError("Test error")

        with pytest.raises(ValueError, match="Test error"):
            with parser.handle_hydra_data(False, dir_path, backup_dir_path, mock_logger):
                raise test_exception

        # verify error is logged
        mock_logger.error.assert_called_with(f"error processing minimal CVE pages: {test_exception}")
        # no restoration debug message should be logged
        assert not any(call.args[0] == "restoring backup directory" for call in mock_logger.debug.call_args_list)

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    @patch('vunnel.providers.rhel.parser.utils.move_dir')
    def test_makedirs_called_correctly(self, mock_move_dir, mock_silent_remove,
                                       mock_exists, mock_makedirs, mock_logger):
        """Test that os.makedirs is called with correct parameters."""
        mock_exists.return_value = False

        dir_path = "/tmp/test/dir"
        backup_dir_path = "/tmp/test/backup"

        with parser.handle_hydra_data(True, dir_path, backup_dir_path, mock_logger):
            pass

        # verify makedirs called for the main directory only (backup should not be created since only move operations will place it)
        expected_calls = [call(dir_path)]
        mock_makedirs.assert_has_calls(expected_calls)
        assert mock_makedirs.call_count == 1

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    @patch('vunnel.providers.rhel.parser.utils.move_dir')
    def test_logging_messages(self, mock_move_dir, mock_silent_remove,
                              mock_exists, mock_makedirs, mock_logger):
        """Test that appropriate logging messages are generated."""
        dir_path = "/custom/path"
        backup_dir_path = "/custom/backup"

        # test with existing directory
        mock_exists.return_value = True
        with parser.handle_hydra_data(True, dir_path, backup_dir_path, mock_logger):
            pass

        mock_logger.debug.assert_called_with(f"moving existing {dir_path} to {backup_dir_path}")

        # reset and test without existing directory
        mock_logger.reset_mock()
        mock_exists.return_value = False

        with parser.handle_hydra_data(True, dir_path, backup_dir_path, mock_logger):
            pass

        mock_logger.debug.assert_called_with(f"no existing {dir_path} to move, starting fresh")

    def test_real_filesystem_operations(self, temp_dirs, mock_logger):
        """Integration test with real filesystem operations."""
        dir_path, backup_dir_path = temp_dirs

        # create initial directory with some content
        os.makedirs(dir_path)
        test_file = os.path.join(dir_path, "test.txt")
        with open(test_file, "w") as f:
            f.write("test content")

        with patch('vunnel.providers.rhel.parser.utils.silent_remove', side_effect=lambda path, tree=False: shutil.rmtree(path) if tree and os.path.exists(path) else None):
            with patch('vunnel.providers.rhel.parser.utils.move_dir', side_effect=shutil.move):
                # test successful execution
                with parser.handle_hydra_data(True, dir_path, backup_dir_path, mock_logger):
                    # directory should exist and be empty
                    assert os.path.exists(dir_path)
                    assert len(os.listdir(dir_path)) == 0

                # recreate the directory with content
                os.makedirs(dir_path, exist_ok=True)
                with open(test_file, "w") as f:
                    f.write("test content")

                # now test exception handling
                with parser.handle_hydra_data(True, dir_path, backup_dir_path, mock_logger):
                    raise ValueError("Test error")

                # directory should be restored with original content
                assert os.path.exists(dir_path)
                assert os.path.exists(test_file)
                with open(test_file, "r") as f:
                    assert f.read() == "test content"

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    @patch('vunnel.providers.rhel.parser.utils.move_dir')
    def test_parameter_combinations(self, mock_move_dir, mock_silent_remove,
                                    mock_exists, mock_makedirs, mock_logger):
        """Test different parameter combinations."""
        test_cases = [
            (True, True),   # ignore_hydra_errors=True, dir exists
            (True, False),  # ignore_hydra_errors=True, dir doesn't exist
            (False, True),  # ignore_hydra_errors=False, dir exists
            (False, False), # ignore_hydra_errors=False, dir doesn't exist
        ]

        for ignore_errors, dir_exists in test_cases:
            mock_exists.return_value = dir_exists
            mock_makedirs.reset_mock()
            mock_silent_remove.reset_mock()
            mock_move_dir.reset_mock()
            mock_logger.reset_mock()

            with parser.handle_hydra_data(ignore_errors, "/tmp/test/dir", "/tmp/test/backup", mock_logger):
                pass

            # all cases should create the main directory
            mock_makedirs.assert_called_with("/tmp/test/dir")

            if ignore_errors:
                # should always remove backup dir and create the main dir (not create the backup dir, since only move operations will place it)
                mock_silent_remove.assert_any_call("/tmp/test/backup", tree=True)
                mock_makedirs.assert_any_call("/tmp/test/dir")

                if dir_exists:
                    # should move existing dir to backup
                    mock_move_dir.assert_called_once_with("/tmp/test/dir", "/tmp/test/backup")
                else:
                    # should not move anything
                    mock_move_dir.assert_not_called()
            else:
                # should only remove main dir
                mock_silent_remove.assert_called_once_with("/tmp/test/dir", tree=True)


    @patch('os.makedirs', side_effect=OSError("Permission denied"))
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    def test_makedirs_failure(self, mock_silent_remove, mock_makedirs, mock_logger):
        """Test behavior when os.makedirs fails."""
        with pytest.raises(OSError, match="Permission denied"):
            with parser.handle_hydra_data(False, "/tmp/test/dir", "/tmp/test/backup", mock_logger):
                pass

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    @patch('vunnel.providers.rhel.parser.utils.move_dir', side_effect=OSError("Move failed"))
    def test_move_dir_failure_during_setup(self, mock_move_dir, mock_silent_remove,
                                           mock_exists, mock_makedirs, mock_logger):
        """Test behavior when vunnel.providers.rhel.parser.utils.move_dir fails during setup."""
        mock_exists.return_value = True

        with pytest.raises(OSError, match="Move failed"):
            with parser.handle_hydra_data(True, "/tmp/test/dir", "/tmp/test/backup", mock_logger):
                pass

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    @patch('vunnel.providers.rhel.parser.utils.move_dir')
    def test_multiple_exceptions(self, mock_move_dir, mock_silent_remove,
                                 mock_exists, mock_makedirs, mock_logger):
        """Test handling of multiple different exception types."""
        mock_exists.return_value = True

        exception_types = [
            ValueError("Value error"),
            RuntimeError("Runtime error"),
            KeyError("Key error"),
            FileNotFoundError("File not found"),
        ]

        for exception in exception_types:
            mock_logger.reset_mock()

            with parser.handle_hydra_data(True, "/tmp/test/dir", "/tmp/test/backup", mock_logger):
                raise exception

            # should log the specific exception
            mock_logger.error.assert_called_with(f"error processing minimal CVE pages: {exception}")

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove', side_effect=OSError("Cleanup failed"))
    @patch('vunnel.providers.rhel.parser.utils.move_dir')
    def test_cleanup_failure_during_exception_handling(self, mock_move_dir, mock_silent_remove,
                                                       mock_exists, mock_makedirs, mock_logger):
        """Test behavior when cleanup operations fail during exception handling."""
        mock_exists.return_value = True

        # the context manager should still handle the original exception properly
        # even if cleanup fails
        with pytest.raises(OSError, match="Cleanup failed"):
            with parser.handle_hydra_data(True, "/tmp/test/dir", "/tmp/test/backup", mock_logger):
                raise ValueError("Original error")

    @patch('os.makedirs')
    @patch('vunnel.providers.rhel.parser.utils.silent_remove')
    def test_empty_paths(self, mock_silent_remove, mock_makedirs, mock_logger):
        """Test behavior with empty or unusual path strings."""
        # test with empty strings
        with parser.handle_hydra_data(False, "", "/tmp/backup", mock_logger):
            pass

        mock_makedirs.assert_called_with("")
        mock_silent_remove.assert_called_with("", tree=True)



from __future__ import annotations

import gzip
import logging
import os
from io import BytesIO
from typing import TYPE_CHECKING

import requests

from vunnel import utils, workspace

if TYPE_CHECKING:
    from collections.abc import Generator


NAMESPACE = "epss"


class Manager:
    _csv_url_template_ = "https://epss.cyentia.com/epss_scores-{}.csv.gz"
    _csv_file_ = "epss_data.csv"

    def __init__(
        self,
        dataset: str,
        ws: workspace.Workspace,
        url_template: str = "",
        download_timeout: int = 125,
        logger: logging.Logger | None = None,
    ):
        self.workspace = ws
        self.download_timeout = download_timeout

        if url_template:
            self._csv_url_template_ = url_template

        self._csv_url_ = self._csv_url_template_.format(dataset)

        self.urls = [self._csv_url_]

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

    def get(self) -> Generator[tuple[str | None, dict[str, str | float]], None, None]:
        filepath = self._download()
        yield from self._normalize(filepath)

    @utils.retry_with_backoff()
    def _download(self) -> str:
        csv_file_path = os.path.join(self.workspace.input_path, self._csv_file_)

        self.logger.info(f"downloading vulnerability data from {self._csv_url_}")

        r = requests.get(self._csv_url_, timeout=self.download_timeout)
        r.raise_for_status()

        with gzip.GzipFile(fileobj=BytesIO(r.content), mode="rb") as gzfh, open(csv_file_path, "wb") as fh:
            fh.write(gzfh.read())

        return csv_file_path

    def _normalize(self, filepath: str) -> Generator[tuple[str, dict[str, str | float]], None, None]:
        fields: list[str] = []
        date: str | None = None

        self.logger.info(f"processing {_count_lines(filepath)} EPSS records")

        with open(filepath, encoding="utf-8") as file_handle:
            for line in file_handle:
                line = line.strip()
                if not line:
                    continue

                if date is None:
                    extracted_date = _extract_date_from_comment(line)
                    if extracted_date:
                        date = extracted_date
                        continue

                if not fields:
                    fields = _parse_header_fields(line)
                    continue

                if not date:
                    raise ValueError("couldn't find date at the top of the CSV file")

                if not line.startswith("CVE-"):
                    continue

                record = _parse_record(line, fields, date)
                if record:
                    cve: str | None = record.get("cve")  # type: ignore[assignment]
                    if not cve:
                        logging.warning(f"skipping record without a CVE: {record}")
                        continue
                    yield cve, record


def _extract_date_from_comment(line: str) -> str | None:
    if not (line.startswith("#") and "score_date" in line):
        return None

    for field in line.split(","):
        if field.startswith("score_date"):
            return field.split(":")[1].split("T")[0]
    return None


def _parse_header_fields(line: str) -> list[str]:
    if line.startswith("CVE-"):
        return []

    fields = [f.lower() for f in line.split(",")]
    if "cve" not in fields:
        raise ValueError("couldn't find 'cve' field in the CSV file header")
    return fields


def _parse_record(line: str, fields: list[str], date: str) -> dict[str, str | float] | None:
    tokens = line.split(",")
    if len(tokens) != len(fields):
        return None

    try:
        input_record = {}
        for i, field in enumerate(fields):
            value = tokens[i]
            input_record[field] = value if field.lower() == "cve" else float(value)
        input_record["date"] = date
        return input_record
    except Exception:
        return None


def _count_lines(filepath: str) -> int:
    with open(filepath) as f:
        return sum(1 for _ in f)



import concurrent.futures
import contextlib
import csv
import functools
import logging
import os
from datetime import UTC, datetime

from vunnel.utils import http_wrapper as http
from vunnel.utils.archive import extract
from vunnel.utils.csaf_types import CSAFDoc
from vunnel.utils.csaf_types import from_path as csaf_from_path
from vunnel.workspace import Workspace

RH_URL_PREFIX = "https://access.redhat.com/errata/"

ADVISORIES_LATEST_URL = "https://security.access.redhat.com/data/csaf/v2/advisories/archive_latest.txt"


class RedHatAdvisoryID:
    def __init__(self, rhsa: str):
        rhsa = rhsa.upper()
        rhsa = rhsa.removeprefix(RH_URL_PREFIX)
        if "-" and ":" in rhsa:
            self.year = rhsa.split("-")[1].split(":")[0]
            self.rhsa = rhsa
        else:
            raise ValueError(f"Invalid RHSA ID: {rhsa}, please provide like RHSA-2021:1234")

    def advisory_url(self) -> str:
        return f"{RH_URL_PREFIX}{self.rhsa}"

    def advisory_year(self) -> str:
        return self.year

    def advisory_id(self) -> str:
        return self.rhsa

    def advisory_kind(self) -> str:
        if self.rhsa.startswith("RHSA"):
            return "security"
        if self.rhsa.startswith("RHBA"):
            return "bugfix"
        if self.rhsa.startswith("RHEA"):
            return "enhancement"
        return "unknown"


class CSAFClient:
    def __init__(
        self,
        workspace: Workspace,
        logger: logging.Logger,
        latest_url: str = ADVISORIES_LATEST_URL,
        skip_download: bool = False,
    ):
        self.workspace = workspace
        self.latest_url = latest_url
        self.latest_filename = "archive_latest.txt"
        self.latest_archive_url: str | None = None
        self.archive_date: datetime | None = None
        self.logger = logger
        self.advisories_path = os.path.join(self.workspace.input_path, "advisories")
        if not skip_download:
            self._download_and_update_archive()
        else:
            self.logger.info("Skipping downloads in RHEL CSAF Client")

    def _changes_url(self) -> str:
        return self.latest_url.replace(self.latest_filename, "changes.csv")

    def _deletions_url(self) -> str:
        return self.latest_url.replace(self.latest_filename, "deletions.csv")

    def _archive_url(self) -> str:
        if not self.latest_archive_url:
            latest_resp = http.get(self.latest_url, logger=self.logger)
            latest_name = latest_resp.text.strip()
            self.latest_archive_url = self.latest_url.replace(self.latest_filename, latest_name)
            date_part = latest_name.removeprefix("csaf_advisories_").removesuffix(".tar.zst")
            self.archive_date = datetime.strptime(date_part, "%Y-%m-%d").replace(tzinfo=UTC)
        return self.latest_archive_url

    def _local_archive_path(self) -> str:
        return os.path.join(self.workspace.input_path, self._archive_url().split("/")[-1])

    def _local_changes_path(self) -> str:
        return os.path.join(self.workspace.input_path, "changes.csv")

    def _local_deletions_path(self) -> str:
        return os.path.join(self.workspace.input_path, "deletions.csv")

    def _download_stream(self, url: str, path: str) -> None:
        with http.get(url, logger=self.logger, stream=True) as response, open(path, "wb") as fh:
            for chunk in response.iter_content(chunk_size=65536):  # 64k chunks
                if chunk:
                    fh.write(chunk)

    def process_changes_and_deletions(self) -> None:
        """process the changes and deletions. deletions.csv is the list of CSAF JSON
        files that have been deleted. Download it and loop over it, deleting all
        referenced files. changes.csv is a date-sorted list of when each CSAF JSON
        file changed. Download it, and loop over the rows, until we get back to the
        date of the archive, keeping a list of unique files, to get the set of files
        that have changed since the archive was published. Re-download all of them, over-writing
        whatever data was in the archive."""
        changes_path = self._local_changes_path()
        deletions_path = self._local_deletions_path()
        with open(deletions_path, newline="") as fh:
            reader = csv.reader(fh)
            for row in reader:
                deleted_fragment = row[0]
                # suppress FileNotFound because deleting the same file twice
                # should no-op rather than raise an error
                with contextlib.suppress(FileNotFoundError):
                    os.remove(os.path.join(self.advisories_path, deleted_fragment))
        seen_files = set()
        years = set()
        with open(changes_path, newline="") as fh:
            reader = csv.reader(fh)
            for row in reader:
                # row is like "2021/cve-2021-47265.json","2024-11-08T18:28:22+00:00"
                changed_file = row[0]
                year = changed_file.split("/")[0]
                date_str = row[1]
                change_date = datetime.fromisoformat(date_str)
                if self.archive_date and change_date < self.archive_date:
                    break
                seen_files.add(changed_file)
                years.add(year)
        # defend against year represented in "changes.csv" but not in main .tar.zst file
        for year in years:
            os.makedirs(os.path.join(self.advisories_path, year), exist_ok=True)
        with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
            futures = {
                executor.submit(
                    self._download_stream,
                    url=self.latest_url.replace("archive_latest.txt", changed_file),
                    path=os.path.join(self.advisories_path, changed_file),
                ): changed_file
                for changed_file in seen_files
            }
            concurrent.futures.wait(futures.keys())
            for future, changed_file in futures.items():
                if future.exception() is not None:
                    self.logger.warning(f"Failed to download {changed_file}: {future.exception()}")

    def _download_and_update_archive(self) -> None:
        if not self.latest_archive_url:
            self.latest_archive_url = self._archive_url()
        archive_path = self._local_archive_path()
        if not os.path.exists(self.advisories_path):
            os.makedirs(self.advisories_path, exist_ok=True)
        # if there's a new one, the paths won't match and we need to download it
        if not os.path.exists(archive_path):
            self._download_stream(self._archive_url(), archive_path)
            extract(archive_path, self.advisories_path)
        # always download and process changes and deletions
        # because this is how the API tells us about updates between
        # publicatins of the main archive.
        self._download_stream(self._changes_url(), os.path.join(self.workspace.input_path, "changes.csv"))
        self._download_stream(self._deletions_url(), os.path.join(self.workspace.input_path, "deletions.csv"))
        self.process_changes_and_deletions()

    def path_from_rhsa_id(self, rhsa_id: RedHatAdvisoryID) -> str:
        # https://security.access.redhat.com/data/csaf/v2/advisories/2024/rhba-2024_0599.json
        return os.path.join(self.advisories_path, rhsa_id.advisory_year(), rhsa_id.advisory_id().lower().replace(":", "_") + ".json")

    def csaf_doc_for_rhsa(self, rhsa: str) -> CSAFDoc | None:
        """Get the CSAF document for a given RHSA ID"""
        doc_path = self.path_from_rhsa_id(RedHatAdvisoryID(rhsa))
        if os.path.exists(doc_path):
            return _csaf_doc_from_path(doc_path)
        self.logger.info(f"CSAF document not found for {rhsa}")
        return None


# this caching is important, because the main RHEL Parser tends to repeatedly
# ask questions about the same RHSA, and we don't want to re-parse the same
# CSAF document over and over.
@functools.lru_cache(maxsize=1024)
def _csaf_doc_from_path(doc_path: str) -> CSAFDoc:
    return csaf_from_path(doc_path)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from vunnel import provider, result, schema

from .manager import Manager

if TYPE_CHECKING:
    import datetime


@dataclass
class Config:
    url: str = "https://www.cisa.gov/sites/default/files/feeds/known_exploited_vulnerabilities.json"
    runtime: provider.RuntimeConfig = field(
        default_factory=lambda: provider.RuntimeConfig(
            result_store=result.StoreStrategy.SQLITE,
            existing_input=provider.InputStatePolicy.DELETE,  # delete the kev.json before running
            existing_results=result.ResultStatePolicy.DELETE_BEFORE_WRITE,
        ),
    )
    request_timeout: int = 125


class Provider(provider.Provider):
    __schema__ = schema.KnownExploitedVulnerabilitySchema()
    __distribution_version__ = int(__schema__.major_version)

    def __init__(self, root: str, config: Config | None = None):
        if not config:
            config = Config()
        super().__init__(root, runtime_cfg=config.runtime)
        self.config = config

        self.logger.debug(f"config: {config}")

        self.manager = Manager(
            url=self.config.url,
            workspace=self.workspace,
            download_timeout=self.config.request_timeout,
            logger=self.logger,
        )

    @classmethod
    def name(cls) -> str:
        return "kev"

    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:
        with self.results_writer() as writer:
            for vuln_id, record in self.manager.get():
                writer.write(
                    identifier=f"kev:{vuln_id.lower()}",
                    schema=self.__schema__,
                    payload=record,
                )

        return self.manager.urls, len(writer)



import logging
import os
from collections.abc import Generator
from pathlib import Path

import orjson

from vunnel.utils import http_wrapper


class Client:
    _default_api_path_ = "/api/v3/osv/"

    def __init__(
        self,
        download_path: str,
        logger: logging.Logger | None = None,
        rocky_versions: list[str] | None = None,
        api_host: str = "https://apollo.build.resf.org",
        skip_download: bool = False,
    ):
        if rocky_versions is None:
            rocky_versions = ["8", "9"]
        self.rocky_versions = rocky_versions
        self.api_host = api_host
        self.urls = [f"{api_host}{self._default_api_path_}"]
        if not logger:
            logger = logging.getLogger("rocky-linux-apollo-client")
        self.logger = logger
        self.download_path = download_path
        self._skip_download = skip_download

    def _download(self) -> None:
        next_page = self._default_api_path_
        while next_page:
            response = http_wrapper.get(f"{self.api_host}{next_page}", self.logger)
            response.raise_for_status()
            data = response.json()
            next_page = data["links"].get("next")
            advisories = data.get("advisories", [])
            for advisory in advisories:
                adv_id = advisory["id"].replace(":", "_")
                write_path = os.path.join(self.download_path, f"{adv_id}.json")
                with open(write_path, "wb") as f:
                    f.write(orjson.dumps(advisory))

    def get(self) -> Generator[Path, None, None]:
        os.makedirs(self.download_path, exist_ok=True)
        if not self._skip_download:
            self._download()
        else:
            self.logger.info("Skipping download of Rocky Linux advisories")
        downloads = Path(self.download_path)
        yield from sorted(downloads.rglob("*.json"))



import logging

from packageurl import PackageURL

from vunnel.providers.rhel.csaf_client import CSAFClient
from vunnel.utils.csaf_types import CSAFDoc
from vunnel.workspace import Workspace


def is_rpm_module_purl(purl: PackageURL) -> bool:
    # Previously RedHat denoted modules with a custom purl type of rpmmod, but now they use an rpmmod
    # qualifier on an rpm purl type, so we need to be able to handle either variation
    if purl.type == "rpmmod":
        return True

    return bool(isinstance(purl.qualifiers, dict) and purl.qualifiers.get("rpmmod"))


def resolve_module_name_from_purl(purl: PackageURL) -> str:
    if purl.type == "rpmmod":
        # The prior redhat module purl looked like pkg:rpmmod/redhat/ruby@2.5:8090020230627084142:b46abd14
        # and we want ruby:2.5 back from that
        mod_version = purl.version or ""
        if mod_version and ":" in mod_version:
            # module versions often look like "2.7:8080020230427102918:63b34585",
            # but we want something like "2.7", so that "modularity: ruby:2.7" can be
            # written into the grype DB.
            mod_version = mod_version.split(":")[0]
        return f"{purl.name}:{mod_version}"

    # The newer format with an rpmmod qualifier on an rpm purl looks like pkg:rpm/redhat/ruby@2.5?rpmmod=ruby:2.5:8090020230627084142:b46abd14,
    # so we can just take the purl components directly
    return f"{purl.name}:{purl.version}"


class CSAFParser:
    """
    CSAFParser is a class that encapsulates transforming CSAF advisory JSON, which is in a broad format
    and requires significant interpretation to make matching data out of it. The primary interface is
    the get_fix_info method, which takes a CVE ID and a parsed affected release object from the Hydra API, and
    returns fixed version and module information for the given CVE, for the package and platform.
    """

    def __init__(self, workspace: Workspace, client: CSAFClient, logger: logging.Logger | None = None, download_timeout: int = 125):
        self.download_timeout = download_timeout
        self.workspace = workspace

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)

        self.logger = logger
        self._urls: set[str] = set()
        self.csaf_client: CSAFClient = client

    @property
    def urls(self) -> list[str]:
        return list(self._urls)

    def platform_module_name_version_from_fpi(self, doc: CSAFDoc, fpi: str) -> tuple[str | None, str | None, str | None, str | None]:
        """
        Given a CSAF document and a full product ID, return the platform, module, name, and version identified
        by the product ID. This essentially de-references a string like
        "AppStream-8.8.0.Z.MAIN.EUS:ruby:2.7:8080020230427102918:63b34585:rubygems-devel-0:3.1.6-139.module+el8.8.0+18745+f1bef313.noarch"
        Into it's component information: this is a RHEL 8 package, from the ruby:2.7 module, and the package is
        is rubygems at version 0:3.1.6-139.module+el8.8.0+18745+f1bef313.

        It would be tempting to simply parse this information out of the string full product ID, but this is challenging because
        There are a variable number of :, since : is used to separate both elements of the product ID and elements of the module
        version, for example, and multiple numbers of -, since - is allowed in package names and is used to separate parts of the
        package version. Therefore, rely on relationships encoded in the document structure to parse this.

        The convention is that a full product ID takes the from {platform}:{module}:{package} where there is
        a module and {platform}:{package} where there is not. The document contains a relationship structure
        that allows unambiguous parsing of the parent child relationships, which is what is employed instead
        of attempting to parse the full product ID only from the single string.
        """

        # The CSAF document only associates purls with ID segments, but associates fixes with full product IDs.
        # That is, given a string like "{platform}:{module}:{package}" or "{platform}:{package}" extract the
        # "{package}" part and ask the CSAF Doc for the associated PURL in order to unambiguously parse a name
        # and version from the PURL.

        # First, get the parents of the product ID.
        # Every full product ID has at least a parent, and possibly a grandparent.
        # If there is no grandparent, the parent is the platform. If there is a grandparent, the parent is the module
        # and the grandparent is the platform.
        module = None
        plat_or_module = doc.product_tree.parent(fpi)
        if not plat_or_module:
            return None, None, None, None
        plat = doc.product_tree.parent(plat_or_module)
        if plat:
            module = plat_or_module.removeprefix(f"{plat}:")
            package = fpi.removeprefix(f"{plat}:{module}:")
        else:
            plat = plat_or_module
            package = fpi.removeprefix(f"{plat}:")

        version = None
        name = None

        purl = doc.product_tree.purl_for_product_id(package)

        if purl:
            parsed_purl = PackageURL.from_string(purl)
            if is_rpm_module_purl(parsed_purl):
                # fpi is a module, not a member of a module; return None; the next pass will find a member of the module if one matches
                return None, None, None, None
            # vunnel fixed versions start with an epoch, which is not part of the PURL version, but instead
            # stored in the PURL qualifiers. If the PURL has an epoch, prepend it to the version.
            epoch = parsed_purl.qualifiers.get("epoch", "0") if isinstance(parsed_purl.qualifiers, dict) else "0"
            version = f"{epoch}:{parsed_purl.version}"
            name = parsed_purl.name
        else:
            self.logger.trace(f"no purl for {package} from {fpi}")  # type: ignore[attr-defined]

        # The CSAF docuemnt product tree has some branches of type product name, and some of type product version.
        # Product version branches are modules and RPMs, and are identified by PURLs. Prduct name branches are platforms,
        # and are identified by CPEs. Iterate the product name branches to find the CPE for the platform component found.
        platform_product_node = next((p for p in doc.product_tree.branches[0].product_name_branches() if p.product_id() == plat), None)
        platform_cpe = platform_product_node.cpe() if platform_product_node else None
        if not platform_cpe:
            self.logger.trace(f"no platform cpe for {plat} from {fpi}")  # type: ignore[attr-defined]
            # this product cannot be attributed to any vunnel namespace, so drop it.
            return None, None, None, None

        if module:
            # If there is a module, get its unambiguous name and version by finding the PURL for the module
            mod_purl = doc.product_tree.purl_for_product_id(module)
            if mod_purl:
                parsed_mod_purl = PackageURL.from_string(mod_purl)
                module = resolve_module_name_from_purl(parsed_mod_purl)
                self.logger.trace(f"module: {module} for {fpi} by {mod_purl}")  # type: ignore[attr-defined]
            else:
                self.logger.trace(f"no module purl for {module} from {fpi}")  # type: ignore[attr-defined]

        # This is enuogh information to compare to an affected release and decide that
        # the patch is about the same package whose vulnerability is mentioned in the CSAF document.
        return platform_cpe, module, name, version

    # note: this is really taking an ar dict because of a circular import issue. It should be taking an AffectedRelease object.
    def get_fix_info(self, cve_id: str, ar: dict[str, str | None], normalized_pkg_name: str) -> tuple[str | None, str | None]:
        """
        Given a CVE ID, an affected release object, and the normalized name of the affected package,
        interrogate CSAF RHSA data to look for an advisory that tells us what version of the package is
        fixed, and the modularity of the package if any.

        The `ar` dict is expected to have the following
        """
        fix_id = ar.get("rhsa_id")
        if not fix_id:
            # This affected release does not reference an advisory; assume it isn't fixed
            return None, None
        doc = self.csaf_client.csaf_doc_for_rhsa(fix_id)
        if not doc:
            # There was no document available for the referenced RHSA. This is unexpected,
            # but logged in the csaf client.
            return None, None

        # The CSAF spec allows N vulnerabilities per document. Find the first one whose CVE matches
        # the CVE ID we're looking for fix info for.
        vuln = next((v for v in doc.vulnerabilities if v.cve == cve_id), None)
        if not vuln:
            self.logger.trace(f"{cve_id}: {fix_id} CSAF doc does not claim to fix this CVE")  # type: ignore[attr-defined]
            return None, None

        # There are multiple remediations in a CSAF doc, including ones that say things like "no fix available".
        # Choose the "vendor_fix" type remediation that corresponds to the advisory ID mentioned in the affected release.
        remediation = next((r for r in vuln.remediations if r.category == "vendor_fix" and r.url and r.url.endswith(fix_id)), None)
        if not remediation:
            self.logger.trace(f"{cve_id} no remediation obj for {fix_id}")  # type: ignore[attr-defined]
            return None, None

        # assume that one of the products listed as rebuilt due to the advisory indicates the first fixed
        # version of the package.
        candidate_full_product_ids = remediation.product_ids
        ar_plat_cpe = ar.get("platform_cpe")
        if not ar_plat_cpe:
            self.logger.trace(f"{cve_id} no platform cpe for {fix_id}")  # type: ignore[attr-defined]
            return None, None
        self.logger.trace(f"{cve_id} searching {fix_id} based on {ar_plat_cpe} and {normalized_pkg_name}")  # type: ignore[attr-defined]

        return self.best_version_module_from_fpis(doc, fix_id, candidate_full_product_ids, normalized_pkg_name, ar_plat_cpe)

    def best_version_module_from_fpis(
        self,
        doc: CSAFDoc,
        fix_id: str,
        candidate_product_ids: list[str],
        normalized_pkg_name: str,
        ar_plat_cpe: str,
    ) -> tuple[str | None, str | None]:
        """
        Loop over all candidate product IDs, that is, all product IDs that were patched by the RHSA in question,
        and ask the CSAF document for their platform, module, name, and version.
        Return the first version and module for the first product ID whose package name matches the
        normalized package name from Hydra, and whose platform CPE starts with the platform CPE from Hydra.
        """

        backup_module, backup_version = None, None
        for fpi in candidate_product_ids:
            plat, module, name, version = self.platform_module_name_version_from_fpi(doc, fpi)
            # use startswith because we see things like cpe:/a:redhat:enterprise_linux:8 in hydra returns
            # and cpe:/o:redhat:enterprise_linux:8::appstream in CSAF docs.
            if name == normalized_pkg_name and plat and plat.startswith(ar_plat_cpe):
                # The names match and the platforms line up, so this is possibly a match.
                # However, verions like like "0:1.4.7-3.ael7b_1" are "alternative" and we should only return
                # them if we can't find a better match.
                if version and "ael" in version:
                    self.logger.trace(f"found alternative match for {fpi}, {name}, {plat} against {ar_plat_cpe}: {normalized_pkg_name}")  # type: ignore[attr-defined]
                    backup_module, backup_version = module, version
                    continue
                self.logger.trace(f"found match for {fpi}, {name}, {plat} against {ar_plat_cpe}: {normalized_pkg_name}")  # type: ignore[attr-defined]
                return version, module
        if backup_version:
            self.logger.trace(  # type: ignore[attr-defined]
                f"returning alternative match {backup_version} (module: {backup_module}) for {fix_id} against {ar_plat_cpe}: {normalized_pkg_name}",
            )
        return backup_version, backup_module



from __future__ import annotations

import os
import shutil
import unittest

import pytest
from vunnel.providers.ubuntu.git import GitRevision, GitWrapper


class TestGitWrapper(unittest.TestCase):
    """
    Unit tests for parsing utilities in GitWrapper

    """

    _workspace_ = "/tmp/ubuntu"
    _location_ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__), "test-fixtures"))

    _git_change_log_file_ = os.path.join(_location_, "git_commits_log")
    _commit_changes_ = [
        {
            "sha": "62187d18d026f17e05af808d95a8c85b7211dda2",
            "updated": {"CVE-2018-14625": "active/CVE-2018-14625"},
            "deleted": {},
        },
        {
            "sha": "eaf213c7cb9118a9bfb59e9b4b736d2d99808ae8",
            "updated": {
                "CVE-2013-7447": "active/CVE-2013-7447",
                "CVE-2017-14650": "retired/CVE-2017-14650",
            },
            "deleted": {},
        },
        {
            "sha": "c182f8269c5bad4201eea14cc185b8ba080374b2",
            "updated": {"CVE-2018-19655": "active/CVE-2018-19655"},
            "deleted": {"CVE-2017-16882": "retired/CVE-2017-16882"},
        },
        {
            "sha": "93c48181efd3a9f220b56715f68a550664425540",
            "updated": {},
            "deleted": {"CVE-2016-8641": "retired/CVE-2016-8641"},
        },
        {
            "sha": "78d22044df36a623b2513bf0e6eaf419c25aed50",
            "updated": {
                "CVE-2017-9773": "active/CVE-2017-9773",
                "CVE-2017-16882": "retired/CVE-2017-16882",
            },
            "deleted": {"CVE-2017-14650": "active/CVE-2017-14650"},
        },
    ]
    _overall_changes_ = {
        "modified": {
            "CVE-2018-14625": "active/CVE-2018-14625",
            "CVE-2013-7447": "active/CVE-2013-7447",
            "CVE-2017-14650": "retired/CVE-2017-14650",
            "CVE-2018-19655": "active/CVE-2018-19655",
            "CVE-2017-9773": "active/CVE-2017-9773",
        },
        "removed": {
            "CVE-2017-16882": "retired/CVE-2017-16882",
            "CVE-2016-8641": "retired/CVE-2016-8641",
        },
    }

    _git_rev_log_file_ = os.path.join(_location_, "git_rev_log")
    _rev_log_cve_ = "CVE-2017-15386"
    _revisions_ = [
        {
            "sha": "e061c9cee652f3347d679ca00b9e9c6c6033994e",
            "file": "retired/CVE-2017-15386",
            "status": "R092",
        },
        {
            "sha": "345fc246b38cb7ffae3e3faff368fbe1de45c6fc",
            "file": "active/CVE-2017-15386",
            "status": "M",
        },
        {
            "sha": "9fb14c9c98f6701dcdc860355f83f8cea0ad0222",
            "file": "active/CVE-2017-15386",
            "status": "A",
        },
        {
            "sha": "2ab08fe597f4b56d4c505773061e1951f04c04cf",
            "file": "active/CVE-2017-15386",
            "status": "M",
        },
        {
            "sha": "0059baff393020a6272adef1ddcd579b91b84e4f",
            "file": "active/CVE-2017-15386",
            "status": "M",
        },
        {
            "sha": "c1dfef873cbc9d089ccefe890dfc645efebedc8c",
            "file": "active/CVE-2017-15386",
            "status": "M",
        },
        {
            "sha": "8845fdd4ec5f551f20f3d74528f9954486ecfa09",
            "file": "active/CVE-2017-15386",
            "status": "M",
        },
        {
            "sha": "7c53b57d6916b89338b1f30f23c3ab47c95b09ee",
            "file": "active/CVE-2017-15386",
            "status": "M",
        },
        {
            "sha": "fec6a993b1d0a7c42cda1925165f922d4dfe9bc4",
            "file": "active/CVE-2017-15386",
            "status": "M",
        },
        {
            "sha": "eb10f42bc38d0fc4bc823472999ad4e129c0ae4b",
            "file": "active/CVE-2017-15386",
            "status": "C056",
        },
    ]

    @classmethod
    def setUpClass(cls):
        os.makedirs(cls._workspace_, exist_ok=True)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls._workspace_)

    def test_parse_log(self):
        with open(self._git_change_log_file_) as f:
            git_commit_log = f.read()

        wrapper = GitWrapper(self._workspace_, "master", self._workspace_)

        commits = wrapper._parse_log(git_commit_log)

        assert len(commits) == len(self._commit_changes_)

        for got, expected in zip(commits, self._commit_changes_):
            assert got.sha in expected["sha"]
            assert got.updated == expected["updated"]
            assert got.deleted == expected["deleted"]
            print(got)

    def test_compute_change_set(self):
        with open(self._git_change_log_file_) as f:
            git_commit_log = f.read()

        wrapper = GitWrapper(self._workspace_, "master", self._workspace_)

        commits = wrapper._parse_log(git_commit_log)

        modified, removed = wrapper._compute_change_set(commits)

        assert modified == self._overall_changes_["modified"]
        assert removed == self._overall_changes_["removed"]
        print(f"Modified: {modified}")
        print(f"Removed: {removed}")


@pytest.mark.parametrize(
    ("git_log_output", "expected"),
    [
        pytest.param(
            """
567e50a45a53a6c32120926656a9b2af0eb10c5b (HEAD -> master, origin/master, origin/HEAD) Process cves run: triaged 2 CVEs, 58 Ignored, 2 Packages
A       active/CVE-2018-14628
A       active/CVE-2022-47630
721abff41afecaa0a499ea49ecaf6c11ac879679 Updating openssl status for trusty/xenial
M       active/CVE-2022-3996
721abff41afecaa0a499ea49ecaf6c11ac879678 update retired
M       retired/CVE-2013-4348
721abff41afecaa0a499ea49ecaf6c11ac879677 move to retired
R100       active/CVE-2013-4348    retired/CVE-2013-4348
721abff41afecaa0a499ea49ecaf6c11ac879676 update
M       active/CVE-2013-4348
721abff41afecaa0a499ea49ecaf6c11ac879675 added
A       active/CVE-2013-4348
    """,
            {
                "CVE-2018-14628": [GitRevision("567e50a45a53a6c32120926656a9b2af0eb10c5b", "active/CVE-2018-14628")],
                "CVE-2022-47630": [GitRevision("567e50a45a53a6c32120926656a9b2af0eb10c5b", "active/CVE-2022-47630")],
                "CVE-2022-3996": [GitRevision("721abff41afecaa0a499ea49ecaf6c11ac879679", "active/CVE-2022-3996")],
                "CVE-2013-4348": [
                    GitRevision("721abff41afecaa0a499ea49ecaf6c11ac879678", "retired/CVE-2013-4348"),
                    GitRevision("721abff41afecaa0a499ea49ecaf6c11ac879676", "active/CVE-2013-4348"),
                    GitRevision("721abff41afecaa0a499ea49ecaf6c11ac879675", "active/CVE-2013-4348"),
                ],
            },
            id="active-retired-active transition",
        ),
    ],
)
def test_parse_full_cve_revision_history(git_log_output: str, expected: dict[str, list[GitRevision]]):
    assert GitWrapper("", "master", "").parse_full_cve_revision_history(git_log_output) == expected



from __future__ import annotations

import copy
import logging
import os
from urllib.parse import urlparse

import orjson

from vunnel.utils import http_wrapper as http
from vunnel.utils import vulnerability


class Parser:
    _release_ = "rolling"
    _secdb_dir_ = "secdb"
    _security_reference_url_ = "https://images.chainguard.dev/security"

    def __init__(  # noqa: PLR0913
        self,
        workspace,
        url: str,
        namespace: str,
        download_timeout: int = 125,
        logger: logging.Logger | None = None,
        security_reference_url: str | None = None,
    ):
        self.download_timeout = download_timeout
        self.secdb_dir_path = os.path.join(workspace.input_path, self._secdb_dir_)
        self.metadata_url = url.strip("/") if url else Parser._url_
        self.url = url
        self.namespace = namespace
        self._db_filename = self._extract_filename_from_url(url)
        self.security_reference_url = security_reference_url.strip("/") if security_reference_url else Parser._security_reference_url_

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

    @staticmethod
    def _extract_filename_from_url(url):
        return os.path.basename(urlparse(url).path)

    def build_reference_links(self, vulnerability_id: str) -> list[str]:
        urls = []
        urls.append(f"{self.security_reference_url}/{vulnerability_id}")
        urls.extend(vulnerability.build_reference_links(vulnerability_id))
        return urls

    def _download(self):
        """
        Downloads wolfi sec db files
        :return:
        """
        if not os.path.exists(self.secdb_dir_path):
            os.makedirs(self.secdb_dir_path, exist_ok=True)

        try:
            self.logger.info(f"downloading {self.namespace} secdb {self.url}")
            r = http.get(self.url, self.logger, stream=True, timeout=self.download_timeout)
            file_path = os.path.join(self.secdb_dir_path, self._db_filename)
            with open(file_path, "wb") as fp:
                for chunk in r.iter_content():
                    fp.write(chunk)
        except Exception:
            self.logger.exception(f"ignoring error processing secdb for {self.url}")

    def _load(self):
        """
        Loads all db json and yields it
        :return:
        """
        dbtype_data_dict = {}

        # parse and transform the json
        try:
            with open(f"{self.secdb_dir_path}/{self._db_filename}") as fh:
                dbtype_data_dict = orjson.loads(fh.read())

                yield self._release_, dbtype_data_dict
        except Exception:
            self.logger.exception(f"failed to load {self.namespace} sec db data")
            raise

    def _normalize(self, release, data):
        """
        Normalize all the sec db entries into vulnerability payload records
        :param release:
        :param dbtype_data_dict:
        :return:
        """

        vuln_dict = {}

        self.logger.debug("normalizing vulnerability data")

        for el in data["packages"]:
            pkg_el = el["pkg"]

            pkg = pkg_el["name"]
            for pkg_version in pkg_el["secfixes"]:
                vids = []
                if pkg_el["secfixes"][pkg_version]:
                    for rawvid in pkg_el["secfixes"][pkg_version]:
                        tmp = rawvid.split()
                        for newvid in tmp:
                            if newvid not in vids:
                                vids.append(newvid)

                for vid in vids:
                    if vid not in vuln_dict:
                        # create a new record
                        vuln_dict[vid] = copy.deepcopy(vulnerability.vulnerability_element)
                        vuln_record = vuln_dict[vid]
                        reference_links = self.build_reference_links(vid)

                        # populate the static information about the new vuln record
                        vuln_record["Vulnerability"]["Name"] = str(vid)
                        vuln_record["Vulnerability"]["NamespaceName"] = self.namespace + ":" + str(release)

                        if reference_links:
                            # TODO: Support multiple links
                            vuln_record["Vulnerability"]["Link"] = reference_links[0]

                        vuln_record["Vulnerability"]["Severity"] = "Unknown"
                    else:
                        vuln_record = vuln_dict[vid]

                    # SET UP fixedins
                    fixed_el = {
                        "Name": pkg,
                        "Version": pkg_version,
                        "VersionFormat": "apk",
                        "NamespaceName": self.namespace + ":" + str(release),
                    }

                    vuln_record["Vulnerability"]["FixedIn"].append(fixed_el)

        return vuln_dict

    def get(self):
        """
        Download, load and normalize wolfi sec db and return a dict of release - list of vulnerability records
        :return:
        """
        # download the data
        self._download()

        # load the data
        for release, dbtype_data_dict in self._load():
            # normalize the loaded data
            yield release, self._normalize(release, dbtype_data_dict)
